# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Benjamin Webb <benjamin.miller.webb@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2023 Benjamin Webb
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
import os
import json
import logging
from typing import Tuple

from dateutil.parser import parse as parse_date
from jsonpatch import make_patch
from jsonschema.exceptions import ValidationError

from pygeoapi.api import API, APIRequest
from pygeoapi.config import get_config, validate_config
from pygeoapi.formats import F_HTML
from pygeoapi.openapi import get_oas
from pygeoapi.util import to_json, render_j2_template, yaml_dump


LOGGER = logging.getLogger(__name__)


class Admin(API):
    """Admin object"""

    PYGEOAPI_CONFIG = os.environ.get('PYGEOAPI_CONFIG')
    PYGEOAPI_OPENAPI = os.environ.get('PYGEOAPI_OPENAPI')

    def __init__(self, config: dict, openapi: dict):
        """
        constructor

        :param config: configuration dict
        :param openapi: openapi dict

        :returns: `pygeoapi.Admin` instance
        """

        super().__init__(config, openapi)

    def merge(self, obj1: dict, obj2: dict):
        """
        Merge two dictionaries

        :param obj1: `dict` of first object
        :param obj2: `dict` of second object

        :returns: `dict` of merged objects
        """

        if isinstance(obj1, dict) and isinstance(obj2, dict):
            merged = obj1.copy()
            for key, value in obj2.items():
                if key in merged:
                    merged[key] = self.merge(merged[key], value)
                else:
                    merged[key] = value
            return merged
        elif isinstance(obj1, list) and isinstance(obj2, list):
            return [self.merge(i1, i2) for i1, i2 in zip(obj1, obj2)]
        else:
            return obj2

    def validate(self, config: dict):
        """
        Validate pygeoapi configuration and OpenAPI to file

        :param config: configuration dict
        """

        # validate pygeoapi configuration
        LOGGER.debug('Validating configuration')
        validate_config(config)
        # validate OpenAPI document
        # LOGGER.debug('Validating openapi document')
        # oas = get_oas(config)
        # validate_openapi_document(oas)
        return True

    def write(self, config: dict):
        """
        Write pygeoapi configuration and OpenAPI to file

        :param config: configuration dict
        """

        self.write_config(config)
        self.write_oas(config)

    def write_config(self, config: dict):
        """
        Write pygeoapi configuration file

        :param config: configuration dict
        """

        # validate pygeoapi configuration
        config = deepcopy(config)
        validate_config(config)

        # Preserve env variables
        LOGGER.debug('Reading env variables in configuration')
        raw_conf = json.loads(to_json(get_config(raw=True)))
        conf = json.loads(to_json(get_config()))
        patch = make_patch(conf, raw_conf)

        LOGGER.debug('Merging env variables')
        config = patch.apply(config)

        # write pygeoapi configuration
        LOGGER.debug('Writing pygeoapi configuration')
        yaml_dump(config, self.PYGEOAPI_CONFIG)
        LOGGER.debug('Finished writing pygeoapi configuration')

    def write_oas(self, config: dict):
        """
        Write pygeoapi OpenAPI document

        :param config: configuration dict
        """

        # validate OpenAPI document
        config = deepcopy(config)
        oas = get_oas(config)
        # validate_openapi_document(oas)

        # write OpenAPI document
        LOGGER.debug('Writing OpenAPI document')
        yaml_dump(oas, self.PYGEOAPI_OPENAPI)
        LOGGER.debug('Finished writing OpenAPI document')


def get_config_(
    admin: Admin,
    request: APIRequest,
) -> Tuple[dict, int, str]:
    """
    Provide admin configuration document

    :param request: request object

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers()

    cfg = get_config(raw=True)

    if request.format == F_HTML:
        content = render_j2_template(
            admin.tpl_config, admin.config['server']['templates'],
            'admin/index.html', cfg, request.locale
        )
    else:
        content = to_json(cfg, admin.pretty_print)

    return headers, 200, content


def put_config(
    admin: Admin,
    request: APIRequest,
) -> Tuple[dict, int, str]:
    """
    Update complete pygeoapi configuration

    :param request: request object

    :returns: tuple of headers, status code, content
    """

    LOGGER.debug('Updating configuration')

    headers = request.get_response_headers()

    data = request.data
    if not data:
        msg = 'missing request data'
        return admin.get_exception(
            400, headers, request.format, 'MissingParameterValue', msg
        )

    try:
        # Parse data
        data = data.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        data = json.loads(data)
        for key, value in data.get('resources', {}).items():
            temporal_extents_str2datetime(value.get('extents', {}))
    except (json.decoder.JSONDecodeError, TypeError) as err:
        # Input is not valid JSON
        LOGGER.error(err)
        msg = 'invalid request data'
        return admin.get_exception(
            400, headers, request.format, 'InvalidParameterValue', msg
        )

    LOGGER.debug('Updating configuration')
    try:
        admin.validate(data)
    except ValidationError as err:
        LOGGER.error(err)
        msg = 'Schema validation error'
        return admin.get_exception(
            400, headers, request.format, 'ValidationError', msg
        )

    admin.write(data)

    return headers, 204, {}


def patch_config(
    admin: Admin, request: APIRequest,
) -> Tuple[dict, int, str]:
    """
    Update partial pygeoapi configuration

    :param request: request object
    :param resource_id: resource identifier

    :returns: tuple of headers, status code, content
    """

    config = deepcopy(admin.config)
    headers = request.get_response_headers()

    data = request.data
    if not data:
        msg = 'missing request data'
        return admin.get_exception(
            400, headers, request.format, 'MissingParameterValue', msg
        )

    try:
        # Parse data
        data = data.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        data = json.loads(data)
        for key, value in data.get('resources', {}).items():
            temporal_extents_str2datetime(value.get('extents', {}))
    except (json.decoder.JSONDecodeError, TypeError) as err:
        # Input is not valid JSON
        LOGGER.error(err)
        msg = 'invalid request data'
        return admin.get_exception(
            400, headers, request.format, 'InvalidParameterValue', msg
        )

    LOGGER.debug('Merging configuration')
    config = admin.merge(config, data)

    try:
        admin.validate(config)
    except ValidationError as err:
        LOGGER.error(err)
        msg = 'Schema validation error'
        return admin.get_exception(
            400, headers, request.format, 'ValidationError', msg
        )

    admin.write(config)

    content = to_json(config, admin.pretty_print)

    return headers, 204, content


def get_resources(
    admin: Admin, request: APIRequest,
) -> Tuple[dict, int, str]:
    """
    Provide admin document

    :param request: request object

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers()

    cfg = get_config(raw=True)

    if request.format == F_HTML:
        content = render_j2_template(
            admin.tpl_config, admin.config['server']['templates'],
            'admin/index.html', cfg['resources'], request.locale
        )
    else:
        content = to_json(cfg['resources'], admin.pretty_print)

    return headers, 200, content


def post_resource(
    admin: Admin, request: APIRequest,
) -> Tuple[dict, int, str]:
    """
    Add resource configuration

    :param request: request object

    :returns: tuple of headers, status code, content
    """

    config = deepcopy(admin.config)
    headers = request.get_response_headers()

    data = request.data
    if not data:
        msg = 'missing request data'
        return admin.get_exception(
            400, headers, request.format, 'MissingParameterValue', msg
        )

    try:
        # Parse data
        data = data.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        data = json.loads(data)
        res = list(data.keys())[0]
        temporal_extents_str2datetime(data[res].get('extents', {}))
    except (json.decoder.JSONDecodeError, TypeError) as err:
        # Input is not valid JSON
        LOGGER.error(err)
        msg = 'invalid request data'
        return admin.get_exception(
            400, headers, request.format, 'InvalidParameterValue', msg
        )

    resource_id = next(iter(data.keys()))

    if config['resources'].get(resource_id) is not None:
        # Resource already exists
        msg = f'Resource exists: {resource_id}'
        LOGGER.error(msg)
        return admin.get_exception(
            400, headers, request.format, 'NoApplicableCode', msg
        )

    LOGGER.debug(f'Adding resource: {resource_id}')
    config['resources'].update(data)

    try:
        admin.validate(config)
    except ValidationError as err:
        LOGGER.error(err)
        msg = 'Schema validation error'
        return admin.get_exception(
            400, headers, request.format, 'ValidationError', msg
        )

    admin.write(config)

    content = f'Location: /{request.path_info}/{resource_id}'
    LOGGER.debug(f'Success at {content}')

    return headers, 201, content


def get_resource(
    admin: Admin, request: APIRequest, resource_id: str
) -> Tuple[dict, int, str]:
    """
    Get resource configuration

    :param request: request object
    :param resource_id:

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers()

    cfg = get_config(raw=True)

    try:
        resource = cfg['resources'][resource_id]
    except KeyError:
        msg = f'Resource not found: {resource_id}'
        return admin.get_exception(
            400, headers, request.format, 'ResourceNotFound', msg
        )

    if request.format == F_HTML:
        content = render_j2_template(
            admin.tpl_config, admin.config['server']['templates'],
            'admin/index.html', resource, request.locale
        )
    else:
        content = to_json(resource, admin.pretty_print)

    return headers, 200, content


def delete_resource(
    admin: Admin, request: APIRequest, resource_id: str
) -> Tuple[dict, int, str]:
    """
    Delete resource configuration

    :param request: request object
    :param resource_id: resource identifier

    :returns: tuple of headers, status code, content
    """

    config = deepcopy(admin.config)
    headers = request.get_response_headers()

    try:
        LOGGER.debug(f'Removing resource configuration for: {resource_id}')
        config['resources'].pop(resource_id)
    except KeyError:
        msg = f'Resource not found: {resource_id}'
        return admin.get_exception(
            400, headers, request.format, 'ResourceNotFound', msg
        )

    LOGGER.debug('Resource removed, validating and saving configuration')
    try:
        admin.validate(config)
    except ValidationError as err:
        LOGGER.error(err)
        msg = 'Schema validation error'
        return admin.get_exception(
            400, headers, request.format, 'ValidationError', msg
        )

    admin.write(config)

    return headers, 204, {}


def put_resource(
    admin: Admin,
    request: APIRequest,
    resource_id: str,
) -> Tuple[dict, int, str]:
    """
    Update complete resource configuration

    :param request: request object
    :param resource_id: resource identifier

    :returns: tuple of headers, status code, content
    """

    config = deepcopy(admin.config)
    headers = request.get_response_headers()

    try:
        LOGGER.debug('Verifying resource exists')
        config['resources'][resource_id]
    except KeyError:
        msg = f'Resource not found: {resource_id}'
        return admin.get_exception(
            400, headers, request.format, 'ResourceNotFound', msg
        )

    data = request.data
    if not data:
        msg = 'missing request data'
        return admin.get_exception(
            400, headers, request.format, 'MissingParameterValue', msg
        )

    try:
        # Parse data
        data = data.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        data = json.loads(data)
        temporal_extents_str2datetime(data.get('extents', {}))
    except (json.decoder.JSONDecodeError, TypeError) as err:
        # Input is not valid JSON
        LOGGER.error(err)
        msg = 'invalid request data'
        return admin.get_exception(
            400, headers, request.format, 'InvalidParameterValue', msg
        )

    LOGGER.debug(f'Updating resource: {resource_id}')
    config['resources'].update({resource_id: data})
    try:
        admin.validate(config)
    except ValidationError as err:
        LOGGER.error(err)
        msg = 'Schema validation error'
        return admin.get_exception(
            400, headers, request.format, 'ValidationError', msg
        )

    admin.write(config)

    return headers, 204, {}


def patch_resource(
    admin: Admin, request: APIRequest, resource_id: str
) -> Tuple[dict, int, str]:
    """
    Update partial resource configuration

    :param request: request object
    :param resource_id: resource identifier

    :returns: tuple of headers, status code, content
    """

    config = deepcopy(admin.config)
    headers = request.get_response_headers()

    try:
        LOGGER.debug('Verifying resource exists')
        resource = config['resources'][resource_id]
    except KeyError:
        msg = f'Resource not found: {resource_id}'
        return admin.get_exception(
            400, headers, request.format, 'ResourceNotFound', msg
        )

    data = request.data
    if not data:
        msg = 'missing request data'
        return admin.get_exception(
            400, headers, request.format, 'MissingParameterValue', msg
        )

    try:
        # Parse data
        data = data.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    try:
        data = json.loads(data)
        temporal_extents_str2datetime(data.get('extents', {}))
    except (json.decoder.JSONDecodeError, TypeError) as err:
        # Input is not valid JSON
        LOGGER.error(err)
        msg = 'invalid request data'
        return admin.get_exception(
            400, headers, request.format, 'InvalidParameterValue', msg
        )

    LOGGER.debug('Merging resource block')
    data = admin.merge(resource, data)
    LOGGER.debug('Updating resource')
    config['resources'].update({resource_id: data})

    try:
        admin.validate(config)
    except ValidationError as err:
        LOGGER.error(err)
        msg = 'Schema validation error'
        return admin.get_exception(
            400, headers, request.format, 'ValidationError', msg
        )

    admin.write(config)

    content = to_json(resource, admin.pretty_print)

    return headers, 204, content


def temporal_extents_str2datetime(extents: dict) -> None:
    """
    Helper function to coerce datetime strings into datetime objects

    :extents: `dict` of pygeoapi resource extents object

    :returns: `None` (changes made directly)
    """

    try:
        extents['temporal']['begin'] = parse_date(extents['temporal']['begin'])
        extents['temporal']['end'] = parse_date(extents['temporal']['end'])
    except (KeyError, TypeError):
        LOGGER.debug('No temporal extents found')



# =================================================================
#
# Authors: Benjamin Webb <bwebb@lincolninst.edu>
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2022 Benjamin Webb
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
import json
from urllib.parse import urlparse
from sodapy import Socrata
import logging

from pygeoapi.crs import crs_transform
from pygeoapi.provider.base import (BaseProvider, ProviderQueryError,
                                    ProviderConnectionError)
from pygeoapi.util import format_datetime

LOGGER = logging.getLogger(__name__)

FIELD_NAME = 'columns_field_name'
DATA_TYPE = 'columns_datatype'


class SODAServiceProvider(BaseProvider):
    """Socrata Open Data API Provider
    """

    def __init__(self, provider_def):
        """
        SODA Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data, id_field, name set in parent class

        :returns: pygeoapi.provider.socrata.SODAServiceProvider
        """
        LOGGER.debug('Logger SODA Init')

        super().__init__(provider_def)
        self.resource_id = provider_def['resource_id']
        self.token = provider_def.get('token')
        self.geom_field = provider_def.get('geom_field')
        self.url = urlparse(self.data).netloc
        self.client = Socrata(self.url, self.token)
        self.get_fields()

    def get_fields(self):
        """
        Get fields of SODA Provider

        :returns: dict of fields
        """

        if not self._fields:

            try:
                [dataset] = self.client.datasets(ids=[self.resource_id])
                resource = dataset['resource']
            except json.decoder.JSONDecodeError as err:
                LOGGER.error(f'Bad response at {self.data}')
                raise ProviderConnectionError(err)

            fields = self.properties or resource[FIELD_NAME]
            for field in fields:
                idx = resource[FIELD_NAME].index(field)
                self._fields[field] = {'type': resource[DATA_TYPE][idx]}

        return self._fields

    @crs_transform
    def query(self, offset=0, limit=10, resulttype='results',
              bbox=[], datetime_=None, properties=[], sortby=[],
              select_properties=[], skip_geometry=False, q=None, **kwargs):
        """
        SODA query

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)

        :returns: dict of GeoJSON FeatureCollection
        """

        # Default feature collection and request parameters

        params = {
            'content_type': 'geojson',
            'select': self._make_fields(select_properties),
            'where': self._make_where(bbox, datetime_, properties),
        }

        fc = {
            'type': 'FeatureCollection',
            'features': [],
            'numberMatched': self._get_count(params)
        }

        if resulttype == 'hits':
            # Return hits
            LOGGER.debug('Returning hits')
            return fc

        if sortby != []:
            params['order'] = self._make_orderby(sortby)

        params['offset'] = offset
        params['limit'] = limit

        def make_feature(f):
            f['id'] = f['properties'][self.id_field]
            if skip_geometry:
                f['geometry'] = None
            return f

        try:
            LOGGER.debug('Sending query')
            resp = self.client.get(self.resource_id, **params)

            LOGGER.debug('Making features')
            fc['features'] = [make_feature(f) for f in resp['features']]
        except Exception as err:
            msg = f'Provider query error: {err}'
            LOGGER.error(msg)
            raise ProviderQueryError(msg)

        fc['numberReturned'] = len(resp['features'])

        return fc

    @crs_transform
    def get(self, identifier, **kwargs):
        """
        Query SODA by id

        :param identifier: feature id

        :returns: dict of single GeoJSON feature
        """
        params = {
            'content_type': 'geojson',
            'limit': 1,
        }
        properties = [(self.id_field, identifier), ]
        params['where'] = self._make_where(properties=properties)

        # Form URL for GET request
        LOGGER.debug('Sending query')
        fc = self.client.get(self.resource_id, **params)
        f = fc.get('features').pop()
        f['id'] = f['properties'][self.id_field]
        return f

    def _make_fields(self, select_properties=[]):
        """
        Make SODA select clause

        :param select_properties: list of property names

        :returns: SODA query `$select` clause
        """
        if self.properties == [] and select_properties == []:
            return '*'

        if self.properties != [] and select_properties != []:
            outFields = set(self.properties) & set(select_properties)
        else:
            outFields = set(self.properties) | set(select_properties)

        outFields = set([self.id_field, *outFields])
        return ','.join(outFields)

    @staticmethod
    def _make_orderby(sortby=[]):
        """
        Make SODA order clause

        :param sortby: `list` of dicts (property, order)

        :returns: SODA query `$order` clause
        """
        __ = {'+': 'ASC', '-': 'DESC'}
        ret = [f"{_['property']} {__[_['order']]}" for _ in sortby]

        return ','.join(ret)

    def _make_where(self, bbox=[], datetime_=None, properties=[]):
        """
        Private function: Make SODA filter from query properties

        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: `list` of tuples (name, value)

        :returns: SODA query `$where` clause
        """

        ret = []

        if properties != []:
            ret.extend(
                [f'{k} = "{v}"' for (k, v) in properties]
            )

        if bbox != []:
            minx, miny, maxx, maxy = bbox
            bpoly = f"'POLYGON (({minx} {miny}, {maxx} {miny}, \
                     {maxx} {maxy}, {minx} {maxy}, {minx} {miny}))'"
            ret.append(f"within_polygon({self.geom_field}, {bpoly})")

        if datetime_ is not None:

            fmt_ = '%Y-%m-%dT%H:%M:%S'
            if '/' in datetime_:
                time_start, time_end = datetime_.split('/')
                if time_start != '..':
                    iso_time = format_datetime(time_start, fmt_)
                    ret.append(f"{self.time_field} >= '{iso_time}'")
                if time_end != '..':
                    iso_time = format_datetime(time_end, fmt_)
                    ret.append(f"{self.time_field} <= '{iso_time}'")

            else:
                iso_time = format_datetime(datetime_, fmt_)
                ret.append(f"{self.time_field} = '{iso_time}'")

        return ' AND '.join(ret)

    def _get_count(self, params):
        """
        Count number of features from query args

        :param params: `dict` of query params

        :returns: `int` of feature count
        """
        params = deepcopy(params)

        params['select'] = 'count(*)'
        params['content_type'] = 'json'

        [response] = self.client.get(self.resource_id, **params)
        return int(response['count'])

    def __repr__(self):
        return f'<SODAServiceProvider> {self.data}'



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from collections import ChainMap
from copy import deepcopy
from datetime import datetime, UTC
from http import HTTPStatus
import logging
from typing import Any, Tuple, Union
import urllib.parse

from pygeofilter.parsers.ecql import parse as parse_ecql_text
from pygeofilter.parsers.cql2_json import parse as parse_cql2_json
from pyproj.exceptions import CRSError

from pygeoapi import l10n
from pygeoapi.api import evaluate_limit
from pygeoapi.api.pubsub import publish_message
from pygeoapi.crs import (DEFAULT_CRS, DEFAULT_STORAGE_CRS,
                          create_crs_transform_spec, get_supported_crs_list,
                          modify_pygeofilter, transform_bbox,
                          set_content_crs_header)
from pygeoapi.formats import F_JSON, FORMAT_TYPES, F_HTML, F_JSONLD
from pygeoapi.formatter.base import FormatterSerializationError
from pygeoapi.linked_data import geojson2jsonld
from pygeoapi.openapi import get_oas_30_parameters
from pygeoapi.plugin import load_plugin, PLUGINS
from pygeoapi.provider import filter_providers_by_type, get_provider_by_type
from pygeoapi.provider.base import (
    ProviderGenericError, ProviderItemNotFoundError,
    ProviderTypeError, SchemaType)

from pygeoapi.util import (to_json, filter_dict_by_key_value, str2bool,
                           render_j2_template, get_dataset_formatters)

from . import APIRequest, API, SYSTEM_LOCALE, validate_bbox, validate_datetime

LOGGER = logging.getLogger(__name__)

OGC_RELTYPES_BASE = 'http://www.opengis.net/def/rel/ogc/1.0'


CONFORMANCE_CLASSES_FEATURES = [
    'http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/core',
    'http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/oas30',
    'http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/html',
    'http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/geojson',
    'http://www.opengis.net/spec/ogcapi-features-2/1.0/conf/crs',
    'http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables',
    'http://www.opengis.net/spec/ogcapi-features-3/1.0/conf/queryables-query-parameters',  # noqa
    'http://www.opengis.net/spec/ogcapi-features-4/1.0/conf/create-replace-delete',  # noqa
    'http://www.opengis.net/spec/ogcapi-features-5/1.0/conf/schemas',
    'http://www.opengis.net/spec/ogcapi-features-5/1.0/conf/core-roles-features'  # noqa
]

CONFORMANCE_CLASSES_RECORDS = [
    'http://www.opengis.net/spec/ogcapi-records-1/1.0/conf/core',
    'http://www.opengis.net/spec/ogcapi-records-1/1.0/conf/sorting',
    'http://www.opengis.net/spec/ogcapi-records-1/1.0/conf/opensearch',
    'http://www.opengis.net/spec/ogcapi-records-1/1.0/conf/json',
    'http://www.opengis.net/spec/ogcapi-records-1/1.0/conf/html'
]


def get_collection_queryables(api: API, request: Union[APIRequest, Any],
                              dataset: str | None = None
                              ) -> Tuple[dict, int, str]:
    """
    Provide collection queryables

    :param request: A request object
    :param dataset: name of collection

    :returns: tuple of headers, status code, content
    """

    domains = {}
    headers = request.get_response_headers(**api.api_headers)

    if any([dataset is None,
            dataset not in api.config['resources'].keys()]):

        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    LOGGER.debug('Creating collection queryables')

    p = None
    for pt in ['feature', 'coverage', 'record']:
        try:
            LOGGER.debug(f'Loading {pt} provider')
            p = load_plugin('provider', get_provider_by_type(
                api.config['resources'][dataset]['providers'], pt))
            break
        except ProviderTypeError:
            LOGGER.debug(f'Providing type {pt} not found')

    if p is None:
        msg = 'queryables not available for this collection'
        return api.get_exception(

            HTTPStatus.BAD_REQUEST, headers, request.format,
            'NoApplicableError', msg)

    LOGGER.debug('Processing profile')
    profile = request.params.get('profile', '')

    LOGGER.debug('Processing properties')
    val = request.params.get('properties')
    if val is not None:
        properties = [x for x in val.split(',') if x]
        properties_to_check = set(p.properties) | set(p.fields.keys())

        if len(list(set(properties) - set(properties_to_check))) > 0:
            msg = 'unknown properties specified'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    else:
        properties = []

    queryables_id = f'{api.get_collections_url()}/{dataset}/queryables'

    if request.params:
        queryables_id += '?' + urllib.parse.urlencode(request.params)

    queryables = {
        'type': 'object',
        'title': l10n.translate(
            api.config['resources'][dataset]['title'], request.locale),
        'properties': {},
        '$schema': 'http://json-schema.org/draft/2019-09/schema',
        '$id': queryables_id
    }

    if p.fields:
        queryables['properties']['geometry'] = {
            'format': 'geometry-any',
            'x-ogc-role': 'primary-geometry'
        }

    if profile == 'actual-domain':
        try:
            domains, _ = p.get_domains(properties)
        except NotImplementedError:
            LOGGER.debug('Domains are not suported by this provider')
            domains = {}

    for k, v in p.fields.items():
        show_field = False
        if properties and k not in properties:
            continue
        if p.properties:
            if k in p.properties:
                show_field = True
        else:
            show_field = True

        if show_field:
            queryables['properties'][k] = {
                'title': k,
                'type': v['type']
            }
            if v['type'] == 'float':
                queryables['properties'][k]['type'] = 'number'
            if v.get('format') is not None:
                queryables['properties'][k]['format'] = v['format']
            if 'values' in v:
                queryables['properties'][k]['enum'] = v['values']

            if k == p.id_field:
                queryables['properties'][k]['x-ogc-role'] = 'id'
            if k == p.time_field:
                queryables['properties'][k]['x-ogc-role'] = 'primary-instant'  # noqa
            if domains.get(k):
                queryables['properties'][k]['enum'] = domains[k]

    if request.format == F_HTML:  # render
        tpl_config = api.get_dataset_templates(dataset)

        queryables['title'] = l10n.translate(
            api.config['resources'][dataset]['title'], request.locale)

        queryables['collections_path'] = api.get_collections_url()
        queryables['dataset_path'] = f'{api.get_collections_url()}/{dataset}'

        content = render_j2_template(api.tpl_config, tpl_config,
                                     'collections/queryables.html',
                                     queryables, request.locale)

        return headers, HTTPStatus.OK, content

    headers['Content-Type'] = 'application/schema+json'

    return headers, HTTPStatus.OK, to_json(queryables, api.pretty_print)


def get_collection_items(
        api: API, request: Union[APIRequest, Any],
        dataset: str | None = None) -> Tuple[dict, int, str]:
    """
    Queries collection

    :param request: A request object
    :param dataset: dataset name

    :returns: tuple of headers, status code, content
    """

    # Set Content-Language to system locale until provider locale
    # has been determined
    headers = request.get_response_headers(SYSTEM_LOCALE,
                                           **api.api_headers)

    properties = []
    reserved_fieldnames = ['bbox', 'bbox-crs', 'crs', 'f', 'lang', 'limit',
                           'offset', 'resulttype', 'datetime', 'sortby',
                           'properties', 'skipGeometry', 'q',
                           'filter', 'filter-lang', 'filter-crs']

    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'collection')

    if dataset not in collections.keys():
        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    LOGGER.debug('Processing query parameters')

    LOGGER.debug('Processing offset parameter')
    try:
        offset = int(request.params.get('offset'))
        if offset < 0:
            msg = 'offset value should be positive or zero'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    except ValueError:
        msg = 'offset value should be an integer'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)
    except TypeError as err:
        LOGGER.warning(err)
        offset = 0

    LOGGER.debug('Processing limit parameter')
    if api.config['server'].get('limit') is not None:
        msg = ('server.limit is no longer supported! '
               'Please use limits at the server or collection '
               'level (RFC5)')
        LOGGER.warning(msg)
    try:
        limit = evaluate_limit(request.params.get('limit'),
                               api.config['server'].get('limits', {}),
                               collections[dataset].get('limits', {}))
    except ValueError as err:
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', str(err))

    resulttype = request.params.get('resulttype') or 'results'

    LOGGER.debug('Processing bbox parameter')

    bbox = request.params.get('bbox')

    if bbox is None:
        bbox = []
    else:
        try:
            bbox = validate_bbox(bbox)
        except ValueError as err:
            msg = str(err)
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)

    LOGGER.debug('Processing datetime parameter')
    datetime_ = request.params.get('datetime')
    try:
        datetime_ = validate_datetime(collections[dataset]['extents'],
                                      datetime_)
    except ValueError as err:
        msg = str(err)
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    LOGGER.debug('processing q parameter')
    q = request.params.get('q') or None

    LOGGER.debug('Loading provider')

    provider_def = None
    try:
        provider_type = 'feature'
        provider_def = get_provider_by_type(
            collections[dataset]['providers'], provider_type)
        p = load_plugin('provider', provider_def)
    except ProviderTypeError:
        try:
            provider_type = 'record'
            provider_def = get_provider_by_type(
                collections[dataset]['providers'], provider_type)
            p = load_plugin('provider', provider_def)
        except ProviderTypeError:
            msg = 'Invalid provider type'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'NoApplicableCode', msg)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    LOGGER.debug('Validating requested format')
    dataset_formatters = get_dataset_formatters(collections[dataset])

    if dataset_formatters:
        LOGGER.debug(f'Dataset formatters: {dataset_formatters}')
        request._format = request._get_format(
            request.get_request_headers(request.headers),
            {v.f: v.mimetype for v in dataset_formatters.values()})

        LOGGER.debug(f'Request format: {request.format}')

    if not request.is_valid(dataset_formatters.keys()):
        return api.get_format_exception(request)

    crs_transform_spec = None
    if provider_type == 'feature':
        # crs query parameter is only available for OGC API - Features
        # right now, not for OGC API - Records.
        LOGGER.debug('Processing crs parameter')
        query_crs_uri = request.params.get('crs')
        try:
            crs_transform_spec = create_crs_transform_spec(
                provider_def, query_crs_uri
            )
        except (ValueError, CRSError) as err:
            msg = str(err)
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
        set_content_crs_header(headers, provider_def, query_crs_uri)

    LOGGER.debug('Processing bbox-crs parameter')
    bbox_crs = request.params.get('bbox-crs')
    if bbox_crs is not None:
        # Validate bbox-crs parameter
        if len(bbox) == 0:
            msg = 'bbox-crs specified without bbox parameter'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'NoApplicableCode', msg)

        if len(bbox_crs) == 0:
            msg = 'bbox-crs specified but is empty'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'NoApplicableCode', msg)

        supported_crs_list = get_supported_crs_list(provider_def)
        if bbox_crs not in supported_crs_list:
            msg = f'bbox-crs {bbox_crs} not supported for this collection'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'NoApplicableCode', msg)
    elif len(bbox) > 0:
        # bbox but no bbox-crs param: assume bbox is in default CRS
        bbox_crs = DEFAULT_STORAGE_CRS

    # Transform bbox to storage_crs
    # when bbox-crs different from storage_crs.
    if len(bbox) > 0:
        try:
            # Get a pyproj CRS instance for the Collection's Storage CRS
            storage_crs = provider_def.get('storage_crs', DEFAULT_STORAGE_CRS)

            # Do the (optional) Transform to the Storage CRS
            bbox = transform_bbox(bbox, bbox_crs, storage_crs)
        except CRSError as e:
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'NoApplicableCode', str(e))

    LOGGER.debug('processing property parameters')
    for k, v in request.params.items():
        include_query_param = False
        if k not in reserved_fieldnames:
            if k in list(p.fields.keys()) or p.include_extra_query_parameters:
                include_query_param = True

        if include_query_param:
            LOGGER.debug(f'Including query parameter {k}={v}')
            properties.append((k, v))
        else:
            LOGGER.debug(f'Discarding query parameter {k}={v}')

    LOGGER.debug('processing sort parameter')
    val = request.params.get('sortby')

    if val is not None:
        sortby = []
        sorts = val.split(',')
        for s in sorts:
            prop = s
            order = '+'
            if s and s[0] in ['+', '-']:
                order = s[0]
                prop = s[1:]

            if prop not in p.fields.keys():
                msg = 'bad sortby property'
                return api.get_exception(
                    HTTPStatus.BAD_REQUEST, headers, request.format,
                    'InvalidParameterValue', msg)

            sortby.append({'property': prop, 'order': order})
    else:
        sortby = []

    LOGGER.debug('processing properties parameter')
    val = request.params.get('properties')

    if val is not None:
        select_properties = val.split(',')
        properties_to_check = set(p.properties) | set(p.fields.keys())

        if (len(list(set(select_properties) -
                     set(properties_to_check))) > 0):
            msg = 'unknown properties specified'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    else:
        select_properties = []

    LOGGER.debug('processing skipGeometry parameter')
    val = request.params.get('skipGeometry')
    if val is not None:
        skip_geometry = str2bool(val)
    else:
        skip_geometry = False

    LOGGER.debug('Processing filter-crs parameter')
    filter_crs_uri = request.params.get('filter-crs', DEFAULT_CRS)

    LOGGER.debug('processing filter parameter')
    cql_text = request.params.get('filter')

    if cql_text is not None:
        try:
            filter_ = parse_ecql_text(cql_text)
            filter_ = modify_pygeofilter(
                filter_,
                filter_crs_uri=filter_crs_uri,
                storage_crs_uri=provider_def.get('storage_crs'),
                geometry_column_name=provider_def.get('geom_field'),
            )
        except Exception:
            msg = 'Bad CQL text'
            LOGGER.error(f'{msg}: {cql_text}')
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    elif request.data:
        try:
            request_data = request.data.decode()
            filter_ = parse_cql2_json(request_data)
            filter_ = modify_pygeofilter(
                filter_,
                filter_crs_uri=filter_crs_uri,
                storage_crs_uri=provider_def.get('storage_crs'),
                geometry_column_name=provider_def.get('geom_field'),
            )
        except Exception:
            msg = 'Bad CQL JSON'
            LOGGER.error(f'{msg}: {request_data}')
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    else:
        filter_ = None

    LOGGER.debug('Processing filter-lang parameter')
    filter_lang = request.params.get('filter-lang')
    # Currently only cql-text is handled, but it is optional
    if filter_lang not in [None, 'cql-json', 'cql-text']:
        msg = 'Invalid filter language'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)
    # Get provider locale (if any)
    prv_locale = l10n.get_plugin_locale(provider_def, request.raw_locale)

    LOGGER.debug('Querying provider')
    LOGGER.debug(f'offset: {offset}')
    LOGGER.debug(f'limit: {limit}')
    LOGGER.debug(f'resulttype: {resulttype}')
    LOGGER.debug(f'sortby: {sortby}')
    LOGGER.debug(f'bbox: {bbox}')
    if provider_type == 'feature':
        LOGGER.debug(f'crs: {query_crs_uri}')
    LOGGER.debug(f'datetime: {datetime_}')
    LOGGER.debug(f'properties: {properties}')
    LOGGER.debug(f'select properties: {select_properties}')
    LOGGER.debug(f'skipGeometry: {skip_geometry}')
    LOGGER.debug(f'language: {prv_locale}')
    LOGGER.debug(f'q: {q}')
    LOGGER.debug(f'cql_text: {cql_text}')
    LOGGER.debug(f'filter_: {filter_}')
    LOGGER.debug(f'filter-lang: {filter_lang}')
    LOGGER.debug(f'filter-crs: {filter_crs_uri}')

    try:
        content = p.query(offset=offset, limit=limit,
                          resulttype=resulttype, bbox=bbox,
                          datetime_=datetime_, properties=properties,
                          sortby=sortby, skip_geometry=skip_geometry,
                          select_properties=select_properties,
                          crs_transform_spec=crs_transform_spec,
                          q=q, language=prv_locale, filterq=filter_)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    serialized_query_params = ''
    for k, v in request.params.items():
        if k not in ('f', 'offset'):
            serialized_query_params += '&'
            serialized_query_params += urllib.parse.quote(k, safe='')
            serialized_query_params += '='
            serialized_query_params += urllib.parse.quote(str(v), safe=',')

    if 'links' not in content:
        content['links'] = []

    # TODO: translate titles
    uri = f'{api.get_collections_url()}/{dataset}/items'
    content['links'].extend([{
        'type': 'application/geo+json',
        'rel': request.get_linkrel(F_JSON),
        'title': l10n.translate('This document as GeoJSON', request.locale),
        'href': f'{uri}?f={F_JSON}{serialized_query_params}'
    }, {
        'rel': request.get_linkrel(F_JSONLD),
        'type': FORMAT_TYPES[F_JSONLD],
        'title': l10n.translate('This document as RDF (JSON-LD)', request.locale),  # noqa
        'href': f'{uri}?f={F_JSONLD}{serialized_query_params}'
    }, {
        'type': FORMAT_TYPES[F_HTML],
        'rel': request.get_linkrel(F_HTML),
        'title': l10n.translate('This document as HTML', request.locale),
        'href': f'{uri}?f={F_HTML}{serialized_query_params}'
    }])

    for key, value in dataset_formatters.items():
        content['links'].append({
            'type': value.mimetype,
            'rel': 'alternate',
            'title': f'This document as {key}',
            'href': f'{uri}?f={value.name}{serialized_query_params}'
        })

    next_link = False
    prev_link = False

    if 'next' in [link['rel'] for link in content['links']]:
        LOGGER.debug('Using next link from provider')
    else:
        if content.get('numberMatched', -1) > (limit + offset):
            next_link = True
        elif len(content['features']) == limit:
            next_link = True

        if offset > 0:
            prev_link = True

    print(request.format)
    if prev_link:
        prev = max(0, offset - limit)
        url = f'{uri}?offset={prev}{serialized_query_params}'
        if request.format is not None:
            url = f'{uri}?f={request.format}&offset={prev}{serialized_query_params}'  # noqa

        content['links'].append(
            {
                'type': 'application/geo+json',
                'rel': 'prev',
                'title': l10n.translate('Items (prev)', request.locale),
                'href': url
            })

    if next_link:
        next_ = offset + limit
        next_href = f'{uri}?offset={next_}{serialized_query_params}'
        if request.format is not None:
            next_href = f'{uri}?f={request.format}&offset={next_}{serialized_query_params}'  # noqa
        content['links'].append(
            {
                'type': 'application/geo+json',
                'rel': 'next',
                'title': l10n.translate('Items (next)', request.locale),
                'href': next_href
            })

    content['links'].append(
        {
            'type': FORMAT_TYPES[F_JSON],
            'title': l10n.translate(
                collections[dataset]['title'], request.locale),
            'rel': 'collection',
            'href': '/'.join(uri.split('/')[:-1])
        })

    content['timeStamp'] = datetime.now(UTC).strftime(
        '%Y-%m-%dT%H:%M:%S.%fZ')

    # Set response language to requested provider locale
    # (if it supports language) and/or otherwise the requested pygeoapi
    # locale (or fallback default locale)
    l10n.set_response_language(headers, prv_locale, request.locale)

    if request.format == F_HTML:  # render
        tpl_config = api.get_dataset_templates(dataset)
        # For constructing proper URIs to items

        content['itemtype'] = p.type
        content['items_path'] = uri
        content['dataset_path'] = '/'.join(uri.split('/')[:-1])
        content['collections_path'] = api.get_collections_url()

        content['offset'] = offset

        content['id_field'] = p.id_field
        if p.uri_field is not None:
            content['uri_field'] = p.uri_field
        if p.title_field is not None:
            content['title_field'] = l10n.translate(p.title_field,
                                                    request.locale)
            # If title exists, use it as id in html templates
            content['id_field'] = content['title_field']
        content = render_j2_template(api.tpl_config, tpl_config,
                                     'collections/items/index.html',
                                     content, request.locale)
        return headers, HTTPStatus.OK, content
    elif request.format in [df.f for df in dataset_formatters.values()]:
        formatter = [v for v in dataset_formatters.values() if
                     v.f == request.format][0]

        try:
            content = formatter.write(
                data=content,
                options={
                    'provider_def': get_provider_by_type(
                        collections[dataset]['providers'],
                        'feature')
                }
            )
        except FormatterSerializationError:
            msg = 'Error serializing output'
            return api.get_exception(
                HTTPStatus.INTERNAL_SERVER_ERROR, headers, request.format,
                'NoApplicableCode', msg)

        headers['Content-Type'] = formatter.mimetype

        if formatter.attachment:
            if p.filename is None:
                filename = f'{dataset}.{formatter.extension}'
            else:
                filename = f'{p.filename}'

            cd = f'attachment; filename="{filename}"'
            headers['Content-Disposition'] = cd

        return headers, HTTPStatus.OK, content

    elif request.format == F_JSONLD:
        content = geojson2jsonld(
            api, content, dataset, id_field=(p.uri_field or 'id')
        )

        return headers, HTTPStatus.OK, content

    return headers, HTTPStatus.OK, to_json(content, api.pretty_print)


def manage_collection_item(
        api: API, request: APIRequest,
        action: str, dataset: str,
        identifier: str | None = None) -> Tuple[dict, int, str]:
    """
    Adds an item to a collection

    :param request: A request object
    :param action: an action among 'create', 'update', 'delete', 'options'
    :param dataset: dataset name

    :returns: tuple of headers, status code, content
    """

    if not request.is_valid(PLUGINS['formatter'].keys()):
        return api.get_format_exception(request)

    # Set Content-Language to system locale until provider locale
    # has been determined
    headers = request.get_response_headers(SYSTEM_LOCALE, **api.api_headers)

    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'collection')

    http_status = HTTPStatus.OK
    payload = None

    if dataset not in collections.keys():
        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    LOGGER.debug('Loading provider')
    try:
        provider_def = get_provider_by_type(
            collections[dataset]['providers'], 'feature')
        p = load_plugin('provider', provider_def)
    except ProviderTypeError:
        try:
            provider_def = get_provider_by_type(
                collections[dataset]['providers'], 'record')
            p = load_plugin('provider', provider_def)
        except ProviderTypeError:
            msg = 'Invalid provider type'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)

    if action == 'options':
        headers['Allow'] = 'HEAD, GET'
        if p.editable:
            if identifier is None:
                headers['Allow'] += ', POST'
            else:
                headers['Allow'] += ', PUT, DELETE'
        return headers, HTTPStatus.OK, ''

    if not p.editable:
        msg = 'Collection is not editable'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    if action in ['create', 'update'] and not request.data:
        msg = 'No data found'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    if action == 'create':
        LOGGER.debug('Creating item')
        try:
            payload = request.data
            identifier = p.create(payload)
        except TypeError as err:
            msg = str(err)
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
        except ProviderGenericError as err:
            return api.get_exception(
                err.http_status_code, headers, request.format,
                err.ogc_exception_code, err.message)

        headers['Location'] = f'{api.get_collections_url()}/{dataset}/items/{identifier}'  # noqa

        http_status = HTTPStatus.CREATED

    if action == 'update':
        LOGGER.debug('Updating item')
        try:
            payload = request.data
            _ = p.update(identifier, payload)
        except TypeError as err:
            msg = str(err)
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
        except ProviderGenericError as err:
            return api.get_exception(
                err.http_status_code, headers, request.format,
                err.ogc_exception_code, err.message)

        http_status = HTTPStatus.NO_CONTENT

    if action == 'delete':
        LOGGER.debug('Deleting item')
        try:
            _ = p.get(identifier)
        except ProviderItemNotFoundError as err:
            return api.get_exception(
                err.http_status_code, headers, request.format,
                err.ogc_exception_code, err.message)

        try:
            _ = p.delete(identifier)
        except ProviderGenericError as err:
            return api.get_exception(
                err.http_status_code, headers, request.format,
                err.ogc_exception_code, err.message)

        http_status = HTTPStatus.OK

    if api.pubsub_client is not None:
        LOGGER.debug('Publishing message')
        publish_message(api.pubsub_client, api.base_url, action, dataset,
                        identifier, payload)

    return headers, http_status, ''


def get_collection_item(api: API, request: APIRequest,
                        dataset: str, identifier: str
                        ) -> Tuple[dict, int, str]:
    """
    Get a single collection item

    :param request: A request object
    :param dataset: dataset name
    :param identifier: item identifier

    :returns: tuple of headers, status code, content
    """

    # Set Content-Language to system locale until provider locale
    # has been determined
    headers = request.get_response_headers(SYSTEM_LOCALE, **api.api_headers)

    LOGGER.debug('Processing query parameters')

    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'collection')

    if dataset not in collections.keys():
        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    LOGGER.debug('Loading provider')

    try:
        provider_type = 'feature'
        provider_def = get_provider_by_type(
            collections[dataset]['providers'], provider_type)
        p = load_plugin('provider', provider_def)
    except ProviderTypeError:
        try:
            provider_type = 'record'
            provider_def = get_provider_by_type(
                collections[dataset]['providers'], provider_type)
            p = load_plugin('provider', provider_def)
        except ProviderTypeError:
            msg = 'Invalid provider type'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    crs_transform_spec = None
    if provider_type == 'feature':
        # crs query parameter is only available for OGC API - Features
        # right now, not for OGC API - Records.
        LOGGER.debug('Processing crs parameter')
        query_crs_uri = request.params.get('crs')
        try:
            crs_transform_spec = create_crs_transform_spec(
                provider_def, query_crs_uri
            )
        except (ValueError, CRSError) as err:
            msg = str(err)
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
        set_content_crs_header(headers, provider_def, query_crs_uri)

    # Get provider language (if any)
    prv_locale = l10n.get_plugin_locale(provider_def, request.raw_locale)

    try:
        LOGGER.debug(f'Fetching id {identifier}')
        content = p.get(
            identifier,
            language=prv_locale,
            crs_transform_spec=crs_transform_spec,
        )
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    if content is None:
        msg = 'identifier not found'
        return api.get_exception(HTTPStatus.BAD_REQUEST, headers,
                                 request.format, 'NotFound', msg)

    uri = content['properties'].get(p.uri_field) if p.uri_field else \
        f'{api.get_collections_url()}/{dataset}/items/{identifier}'

    if 'links' not in content:
        content['links'] = []

    content['links'].extend([{
        'type': FORMAT_TYPES[F_JSON],
        'rel': 'root',
        'title': l10n.translate('The landing page of this server as JSON', request.locale),  # noqa
        'href': f"{api.base_url}?f={F_JSON}"
        }, {
        'type': FORMAT_TYPES[F_HTML],
        'rel': 'root',
        'title': l10n.translate('The landing page of this server as HTML', request.locale),  # noqa
        'href': f"{api.base_url}?f={F_HTML}"
        }, {
        'rel': request.get_linkrel(F_JSON),
        'type': 'application/geo+json',
        'title': l10n.translate('This document as JSON', request.locale),
        'href': f'{uri}?f={F_JSON}'
        }, {
        'rel': request.get_linkrel(F_JSONLD),
        'type': FORMAT_TYPES[F_JSONLD],
        'title': l10n.translate('This document as RDF (JSON-LD)', request.locale),  # noqa
        'href': f'{uri}?f={F_JSONLD}'
        }, {
        'rel': request.get_linkrel(F_HTML),
        'type': FORMAT_TYPES[F_HTML],
        'title': l10n.translate('This document as HTML', request.locale),
        'href': f'{uri}?f={F_HTML}'
        }, {
        'rel': 'collection',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate(collections[dataset]['title'],
                                request.locale),
        'href': f'{api.get_collections_url()}/{dataset}'
    }])

    link_request_format = (
        request.format if request.format is not None else F_JSON
    )
    if 'prev' in content:
        content['links'].append({
            'rel': 'prev',
            'type': FORMAT_TYPES[link_request_format],
            'href': f"{api.get_collections_url()}/{dataset}/items/{content['prev']}?f={link_request_format}"  # noqa
        })
    if 'next' in content:
        content['links'].append({
            'rel': 'next',
            'type': FORMAT_TYPES[link_request_format],
            'href': f"{api.get_collections_url()}/{dataset}/items/{content['next']}?f={link_request_format}"  # noqa
        })

    # Set response language to requested provider locale
    # (if it supports language) and/or otherwise the requested pygeoapi
    # locale (or fallback default locale)
    l10n.set_response_language(headers, prv_locale, request.locale)

    if request.format == F_HTML:  # render
        tpl_config = api.get_dataset_templates(dataset)
        content['title'] = l10n.translate(collections[dataset]['title'],
                                          request.locale)
        content['id_field'] = p.id_field
        if p.uri_field is not None:
            content['uri_field'] = p.uri_field
        if p.title_field is not None:
            content['title_field'] = l10n.translate(p.title_field,
                                                    request.locale)
        content['collections_path'] = api.get_collections_url()

        content = render_j2_template(api.tpl_config, tpl_config,
                                     'collections/items/item.html',
                                     content, request.locale)
        return headers, HTTPStatus.OK, content

    elif request.format == F_JSONLD:
        content = geojson2jsonld(
            api, content, dataset, uri, (p.uri_field or 'id')
        )

        return headers, HTTPStatus.OK, content

    return headers, HTTPStatus.OK, to_json(content, api.pretty_print)


def get_oas_30(cfg: dict, locale: str) -> tuple[list[dict[str, str]], dict[str, dict]]:  # noqa
    """
    Get OpenAPI fragments

    :param cfg: `dict` of configuration
    :param locale: `str` of locale

    :returns: `tuple` of `list` of tag objects, and `dict` of path objects
    """

    from pygeoapi.openapi import OPENAPI_YAML, get_visible_collections

    limit = {
        'name': 'limit',
        'in': 'query',
        'description': 'The optional limit parameter limits the number of items that are presented in the response document',  # noqa
        'required': False,
        'schema': {
            'type': 'integer',
            'minimum': 1,
            'maximum': 10000,
            'default': 100
        },
        'style': 'form',
        'explode': False
    }

    profile = {
        'name': 'profile',
        'in': 'query',
        'description': 'The profile to be applied to a given request',
        'required': False,
        'style': 'form',
        'explode': False,
        'schema': {
            'type': 'string',
            'enum': ['actual-domain', 'valid-domain']
        }
    }

    LOGGER.debug('setting up collection endpoints')
    paths = {}

    collections = filter_dict_by_key_value(cfg['resources'],
                                           'type', 'collection')

    for k, v in get_visible_collections(cfg).items():
        try:
            ptype = None

            if filter_providers_by_type(
                    collections[k]['providers'], 'feature'):
                ptype = 'feature'

            if filter_providers_by_type(
                    collections[k]['providers'], 'record'):
                ptype = 'record'

            p = load_plugin('provider', get_provider_by_type(
                            collections[k]['providers'], ptype))

            collection_name_path = f'/collections/{k}'
            items_path = f'/collections/{k}/items'
            title = l10n.translate(v['title'], locale)
            description = l10n.translate(v['description'], locale)

            oas_30_parameters = get_oas_30_parameters(cfg, locale)

            coll_properties = deepcopy(oas_30_parameters)['properties']
            coll_properties['schema']['items']['enum'] = list(p.fields.keys())

            coll_limit = _derive_limit(
                deepcopy(limit), cfg['server'].get('limits', {}),
                v.get('limits', {})
            )

            dataset_formatters = get_dataset_formatters(v)
            coll_f_parameter = deepcopy(oas_30_parameters)['f']
            for key, value in dataset_formatters.items():
                coll_f_parameter['schema']['enum'].append(value.f)

            paths[items_path] = {
                'get': {
                    'summary': f'Get {title} items',
                    'description': description,
                    'tags': [k],
                    'operationId': f'get{k.capitalize()}Features',
                    'parameters': [
                        coll_f_parameter,
                        {'$ref': '#/components/parameters/lang'},
                        {'$ref': '#/components/parameters/bbox'},
                        coll_limit,
                        {'$ref': '#/components/parameters/crs'},  # noqa
                        {'$ref': '#/components/parameters/bbox-crs'},
                        coll_properties,
                        {'$ref': '#/components/parameters/vendorSpecificParameters'},  # noqa
                        {'$ref': '#/components/parameters/skipGeometry'},
                        {'$ref': f"{OPENAPI_YAML['oapir']}/parameters/sortby.yaml"},  # noqa
                        {'$ref': '#/components/parameters/offset'}
                    ],
                    'responses': {
                        '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Features"},  # noqa
                        '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                        '404': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/NotFound"},  # noqa
                        '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                    }
                },
                'options': {
                    'summary': f'Options for {title} items',
                    'description': description,
                    'tags': [k],
                    'operationId': f'options{k.capitalize()}Features',
                    'responses': {
                        '200': {'description': 'options response'}
                    }
                }
            }

            # TODO: update feature POSTs once updated in OGC API - Features
            # https://github.com/opengeospatial/ogcapi-features/issues/771
            paths[items_path]['post'] = {
                'summary': f'Get {title} items with CQL2',
                'description': description,
                'tags': [k],
                'operationId': f'getCQL2{k.capitalize()}Features',
                'requestBody': {
                    'description': 'Get items with CQL2',
                    'content': {
                        'application/json': {  # CQL2
                            'schema': {
                                '$ref': OPENAPI_YAML['cql2']
                            }
                        }
                    },
                    'required': True
                },
                'responses': {
                    '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Features"},  # noqa
                    '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                    '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                }
            }

            if p.editable:
                LOGGER.debug('Provider is editable; adding post')

                paths[items_path]['post']['operationId'] = f'getCQL2OrAdd{k.capitalize()}Features'  # noqa

                val = paths[items_path]['post']['summary']
                paths[items_path]['post']['summary'] = f'{val} or add item (see media type)'  # noqa

                val = paths[items_path]['post']['requestBody']['description']
                paths[items_path]['post']['requestBody']['description'] = f'{val} or add item to collection'  # noqa

                paths[items_path]['post']['requestBody']['content']['application/geo+json'] = {  # noqa
                    'schema': {}
                }

                paths[items_path]['post']['responses']['201'] = {
                    'description': 'Successful creation'
                }

                try:
                    schema_ref = p.get_schema(SchemaType.create)
                    paths[items_path]['post']['requestBody']['content'][schema_ref[0]] = {  # noqa
                        'schema': schema_ref[1]
                    }
                except Exception as err:
                    LOGGER.debug(err)

            if ptype == 'record':
                paths[items_path]['get']['parameters'].append(
                    {'$ref': f"{OPENAPI_YAML['oapir']}/parameters/q.yaml"})
            if p.fields:
                schema_path = f'{collection_name_path}/schema'

                paths[schema_path] = {
                    'get': {
                        'summary': f'Get {title} schema',
                        'description': description,
                        'tags': [k],
                        'operationId': f'get{k.capitalize()}Schema',
                        'parameters': [
                            {'$ref': '#/components/parameters/f'},
                            {'$ref': '#/components/parameters/lang'}
                        ],
                        'responses': {
                            '200': {'$ref': '#/components/responses/Queryables'},  # noqa
                            '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                            '404': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/NotFound"},  # noqa
                            '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"},  # noqa
                        }
                    }
                }

                queryables_path = f'{collection_name_path}/queryables'

                paths[queryables_path] = {
                    'get': {
                        'summary': f'Get {title} queryables',
                        'description': description,
                        'tags': [k],
                        'operationId': f'get{k.capitalize()}Queryables',
                        'parameters': [
                            coll_properties,
                            {'$ref': '#/components/parameters/f'},
                            profile,
                            {'$ref': '#/components/parameters/lang'}
                        ],
                        'responses': {
                            '200': {'$ref': '#/components/responses/Queryables'},  # noqa
                            '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                            '404': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/NotFound"},  # noqa
                            '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"},  # noqa
                        }
                    }
                }

            if p.time_field is not None:
                paths[items_path]['get']['parameters'].append(
                    {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/datetime"})  # noqa

            for field, type_ in p.fields.items():

                if p.properties and field not in p.properties:
                    LOGGER.debug('Provider specified not to advertise property')  # noqa
                    continue

                if field == 'q' and ptype == 'record':
                    LOGGER.debug('q parameter already declared, skipping')
                    continue

                if type_ == 'date':
                    schema = {
                        'type': 'string',
                        'format': 'date'
                    }
                elif type_ == 'float':
                    schema = {
                        'type': 'number',
                        'format': 'float'
                    }
                elif type_ == 'long':
                    schema = {
                        'type': 'integer',
                        'format': 'int64'
                    }
                else:
                    schema = type_

                if schema.get('format') is None:
                    schema.pop('format', None)

                path_ = f'{collection_name_path}/items'
                paths[path_]['get']['parameters'].append({
                    'name': field,
                    'in': 'query',
                    'required': False,
                    'schema': schema,
                    'style': 'form',
                    'explode': False
                })

            paths[f'{collection_name_path}/items/{{featureId}}'] = {
                'get': {
                    'summary': f'Get {title} item by id',
                    'description': description,
                    'tags': [k],
                    'operationId': f'get{k.capitalize()}Feature',
                    'parameters': [
                        {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/featureId"},  # noqa
                        {'$ref': '#/components/parameters/crs'},  # noqa
                        {'$ref': '#/components/parameters/f'},
                        {'$ref': '#/components/parameters/lang'}
                    ],
                    'responses': {
                        '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Feature"},  # noqa
                        '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                        '404': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/NotFound"},  # noqa
                        '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                    }
                },
                'options': {
                    'summary': f'Options for {title} item by id',
                    'description': description,
                    'tags': [k],
                    'operationId': f'options{k.capitalize()}Feature',
                    'parameters': [
                        {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/featureId"}  # noqa
                    ],
                    'responses': {
                        '200': {'description': 'options response'}
                    }
                }
            }

            try:
                schema_ref = p.get_schema()
                paths[f'{collection_name_path}/items/{{featureId}}']['get']['responses']['200'] = {  # noqa
                    'content': {
                        schema_ref[0]: {
                            'schema': schema_ref[1]
                        }
                    }
                }
            except Exception as err:
                LOGGER.debug(err)

            if p.editable:
                LOGGER.debug('Provider is editable; adding put/delete')
                put_path = f'{collection_name_path}/items/{{featureId}}'  # noqa
                paths[put_path]['put'] = {  # noqa
                    'summary': f'Update {title} items',
                    'description': description,
                    'tags': [k],
                    'operationId': f'update{k.capitalize()}Features',
                    'parameters': [
                        {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/featureId"}  # noqa
                    ],
                    'requestBody': {
                        'description': 'Updates item in collection',
                        'content': {
                            'application/geo+json': {
                                'schema': {}
                            }
                        },
                        'required': True
                    },
                    'responses': {
                        '204': {'$ref': '#/components/responses/204'},
                        '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                        '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                    }
                }

                try:
                    schema_ref = p.get_schema(SchemaType.replace)
                    paths[put_path]['put']['requestBody']['content'][schema_ref[0]] = {  # noqa
                        'schema': schema_ref[1]
                    }
                except Exception as err:
                    LOGGER.debug(err)

                paths[f'{collection_name_path}/items/{{featureId}}']['delete'] = {  # noqa
                    'summary': f'Delete {title} items',
                    'description': description,
                    'tags': [k],
                    'operationId': f'delete{k.capitalize()}Features',
                    'parameters': [
                        {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/featureId"},  # noqa
                    ],
                    'responses': {
                        '200': {'description': 'Successful delete'},
                        '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                        '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                    }
                }

        except ProviderTypeError:
            LOGGER.debug('collection is not feature/item based')

    return [{'name': 'records'}, {'name': 'features'}], {'paths': paths}


def _derive_limit(limit_object, server_limits, collection_limits) -> dict:
    """
    Helper function to derive a limit object for a given collection

    :param limit_object: OpenAPI limit parameter
    :param server_limits: server level limits configuration
    :param collection_limits: collection level limits configuration

    :returns: updated limit object
    """

    effective_limits = ChainMap(collection_limits, server_limits)

    default_limit = effective_limits.get('default_items', 10)
    max_limit = effective_limits.get('max_items', 10)

    limit_object['schema']['default'] = default_limit
    limit_object['schema']['maximum'] = max_limit

    text = f' (maximum={max_limit}, default={default_limit}).'
    limit_object['description'] += text

    return limit_object



# =================================================================
#
# Authors: Ben Webb <bwebb@lincolninst.edu>
#
# Copyright (c) 2025 Ben Webb
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from pygeoapi.provider.base import ProviderNoDataError
from pygeoapi.provider.base_edr import BaseEDRProvider
from pygeoapi.provider.sensorthings import SensorThingsProvider

LOGGER = logging.getLogger(__name__)

GEOGRAPHIC_CRS = {
    'coordinates': ['x', 'y'],
    'system': {
        'type': 'GeographicCRS',
        'id': 'http://www.opengis.net/def/crs/OGC/1.3/CRS84'
    }
}

TEMPORAL_RS = {
    'coordinates': ['t'],
    'system': {'type': 'TemporalRS', 'calendar': 'Gregorian'}
}


class SensorThingsEDRProvider(BaseEDRProvider, SensorThingsProvider):
    def __init__(self, provider_def):
        """
        Initialize the SensorThingsEDRProvider instance.

        :param provider_def: provider definition
        """

        provider_def['entity'] = 'ObservedProperties'
        BaseEDRProvider.__init__(self, provider_def)
        SensorThingsProvider.__init__(self, provider_def)
        self.expand['ObservedProperties'] = (
            'Datastreams/Thing/Locations,Datastreams/Observations'
        )
        self.time_field = 'Datastreams/Observations/resultTime'
        self._fields = {}
        self.get_fields()

    def get_fields(self):
        """
        Retrieve and store ObservedProperties field definitions.

        :returns: A dictionary mapping field IDs to their properties.
        """

        if not self._fields:
            r = self._get_response(
                self._url, entity='ObservedProperties', expand='Datastreams'
            )
            try:
                _ = r['value'][0]
            except IndexError:
                LOGGER.warning('could not get fields; returning empty set')
                return {}

            for feature in r['value']:
                id = str(feature['@iot.id'])
                key = feature['name']
                try:
                    uom = feature['Datastreams'][0]['unitOfMeasurement']
                except IndexError:
                    continue

                self._fields[id] = {
                    'type': 'number',
                    'title': key,
                    'x-ogc-unit': uom['symbol']
                }

        return self._fields

    def items(self, **kwargs):
        """
        Retrieve a collection of items.

        :param kwargs: Additional parameters for the request.
        :returns: A GeoJSON representation of the items.
        """

        # This method is empty due to the way pygeoapi handles items requests
        # We implement this method inside of the feature provider
        pass

    def locations(
        self,
        select_properties: list = [],
        bbox: list = [],
        datetime_: str = None,
        location_id: str = None,
        **kwargs
    ):
        """
        Extract and return location data from ObservedProperties.

        :param select_properties: List of properties to include.
        :param bbox: Bounding box geometry for spatial queries.
        :param datetime_: Temporal filter for observations.
        :param location_id: Identifier of the location to filter by.

        :returns: A GeoJSON FeatureCollection of locations.
        """

        fc = {'type': 'FeatureCollection', 'features': []}

        params = {}
        expand = [
            'Datastreams($select=description,name,unitOfMeasurement)',
            'Datastreams/Thing($select=@iot.id)',
            'Datastreams/Thing/Locations($select=location)'
        ]

        if select_properties:
            properties = [
                ['@iot.id', f"'{p}'" if isinstance(p, str) else p]
                for p in select_properties
            ]
            ret = [f'{name} eq {value}' for (name, value) in properties]
            params['$filter'] = ' or '.join(ret)

        filter_ = f'$filter={self._make_dtf(datetime_)};' if datetime_ else ''
        if location_id:
            expand[0] = (
                f'Datastreams($filter=Thing/@iot.id eq {location_id};'
                '$select=description,name,unitOfMeasurement)'
            )
            expand.append(
                f'Datastreams/Observations({filter_}$orderby=phenomenonTime;'
                '$select=result,phenomenonTime,resultTime)'
            )
        else:
            expand.append(
                f'Datastreams/Observations({filter_}$select=result;$top=1)')

        if bbox:
            geom_filter = self._make_bbox(bbox, 'Datastreams')
            expand[0] = f'Datastreams($filter={geom_filter})'

        expand = ','.join(expand)
        response = self._get_response(
            url=self._url, params=params,
            entity='ObservedProperties', expand=expand
        )

        if location_id:
            return self._make_coverage_collection(response)

        for property in response['value']:
            for datastream in property['Datastreams']:
                if len(datastream['Observations']) == 0:
                    continue

                fc['features'].append(
                    self._make_feature(datastream['Thing'], entity='Things')
                )

        return fc

    def cube(
        self,
        select_properties: list = [],
        bbox: list = [],
        datetime_: str = None,
        **kwargs
    ):
        """
        Extract and return coverage data from ObservedProperties.

        :param select_properties: List of properties to include.
        :param bbox: Bounding box geometry for spatial queries.
        :param datetime_: Temporal filter for observations.

        :returns: A CovJSON CoverageCollection.
        """

        params = {}

        geom_filter = self._make_bbox(bbox, 'Datastreams')
        expand = [
            (
             f'Datastreams($filter={geom_filter};'
             '$select=description,name,unitOfMeasurement)'
            ),
            'Datastreams/Thing($select=@iot.id)',
            'Datastreams/Thing/Locations($select=location)'
        ]

        if select_properties:
            properties = [
                ['@iot.id', f"'{p}'" if isinstance(p, str) else p]
                for p in select_properties
            ]
            ret = [f'{name} eq {value}' for (name, value) in properties]
            params['$filter'] = ' or '.join(ret)

        filter_ = f'$filter={self._make_dtf(datetime_)};' if datetime_ else ''
        expand.append(
            f'Datastreams/Observations({filter_}$orderby=phenomenonTime;'
            '$select=result,phenomenonTime,resultTime)'
        )

        expand = ','.join(expand)
        response = self._get_response(
            url=self._url, params=params,
            entity='ObservedProperties', expand=expand
        )

        return self._make_coverage_collection(response)

    def area(
        self,
        wkt: str,
        select_properties: list = [],
        datetime_: str = None,
        **kwargs
    ):
        """
        Extract and return coverage data from a specified area.

        :param wkt: Well-Known Text (WKT) representation of the
                    geometry for the area.
        :param select_properties: List of properties to include
                                  in the response.
        :param datetime_: Temporal filter for observations.

        :returns: A CovJSON CoverageCollection.
        """

        params = {}

        expand = [
            (
             'Datastreams($filter=st_within('
             f"Thing/Locations/location,geography'{wkt}');"
             '$select=description,name,unitOfMeasurement)'
            ),
            'Datastreams/Thing($select=@iot.id)',
            'Datastreams/Thing/Locations($select=location)'
        ]

        if select_properties:
            properties = [
                ['@iot.id', f"'{p}'" if isinstance(p, str) else p]
                for p in select_properties
            ]
            ret = [f'{name} eq {value}' for (name, value) in properties]
            params['$filter'] = ' or '.join(ret)

        filter_ = f'$filter={self._make_dtf(datetime_)};' if datetime_ else ''
        expand.append(
            f'Datastreams/Observations({filter_}$orderby=phenomenonTime;'
            '$select=result,phenomenonTime,resultTime)'
        )

        expand = ','.join(expand)
        response = self._get_response(
            url=self._url, params=params,
            entity='ObservedProperties', expand=expand
        )

        return self._make_coverage_collection(response)

    def _generate_coverage(self, datastream: dict, id: str) -> dict:
        """
        Generate a coverage object for a datastream.

        :param datastream: The datastream data to generate coverage for.
        :param id: The ID to use for the coverage.

        :returns: A dict containing the coverage object.
        """

        times, values = self._expand_observations(datastream)
        thing = datastream['Thing']
        coords = thing['Locations'][0]['location']['coordinates']
        length = len(values)

        return {
            'type': 'Coverage',
            'id': str(thing['@iot.id']),
            'domain': {
                'type': 'Domain',
                'domainType': 'PointSeries',
                'axes': {
                    'x': {'values': [coords[0]]},
                    'y': {'values': [coords[1]]},
                    't': {'values': times}
                },
                'referencing': [GEOGRAPHIC_CRS, TEMPORAL_RS]
            },
            'ranges': {
                id: {
                    'type': 'NdArray',
                    'dataType': 'float',
                    'axisNames': ['t'],
                    'shape': [length],
                    'values': values
                }
            }
        }, length

    @staticmethod
    def _generate_paramters(datastream: dir, label: str) -> dict:
        """
        Generate parameters for a given datastream.

        :param datastream: The datastream data to generate parameters for.
        :param label: The label for the parameter.

        :returns: A dictionary containing the parameter definition.
        """

        return {
            'type': 'Parameter',
            'description': {'en': datastream['description']},
            'observedProperty': {'id': label, 'label': {'en': label}},
            'unit': {
                'label': {'en': datastream['unitOfMeasurement']['name']},
                'symbol': datastream['unitOfMeasurement']['symbol']
            }
        }

    @staticmethod
    def _make_dtf(datetime_: str) -> str:
        """
        Create a Date-Time filter for querying.

        :param datetime_: Temporal filter.

        :returns: A string datetime filter for use in queries.
        """

        dtf_r = []
        if '/' in datetime_:
            time_start, time_end = datetime_.split('/')
            if time_start != '..':
                dtf_r.append(f'phenomenonTime ge {time_start}')

            if time_end != '..':
                dtf_r.append(f'phenomenonTime le {time_end}')

        else:
            dtf_r.append(f'phenomenonTime eq {datetime_}')

        return ' and '.join(dtf_r)

    def _make_coverage_collection(self, response):
        """
        Build a CoverageCollection from the SensorThings API response.

        :param response: source response from SensorThings.

        :returns: The updated CoverageCollection object.
        """

        cc = {
            'type': 'CoverageCollection',
            'domainType': 'PointSeries',
            'parameters': {},
            'coverages': []
        }

        for feature in response['value']:
            if len(feature['Datastreams']) == 0:
                continue

            id = feature['name'].replace(' ', '+')

            for datastream in feature['Datastreams']:
                coverage, length = self._generate_coverage(datastream, id)
                if length > 0:
                    cc['coverages'].append(coverage)
                    cc['parameters'][id] = self._generate_paramters(
                        datastream, feature['name']
                    )

        if cc['parameters'] == {} or cc['coverages'] == []:
            msg = 'No data found'
            LOGGER.warning(msg)
            raise ProviderNoDataError(msg)

        return cc

    @staticmethod
    def _expand_observations(datastream: dict):
        """
        Expand and extract observation times and values from a datastream.

        :param datastream: The datastream containing observations.

        :returns: A tuple containing lists of times and values.
        """

        times = []
        values = []
        # TODO: Expand observations when 'Observations@iot.nextLink'
        # or '@iot.nextLink' is present
        for obs in datastream['Observations']:
            resultTime = obs['resultTime'] or obs['phenomenonTime']
            if obs['result'] is not None and resultTime:
                try:
                    result = float(obs['result'])
                except ValueError:
                    result = obs['result']
                times.append(resultTime)
                values.append(result)

        return times, values



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from collections import OrderedDict

F_JSON = 'json'
F_COVERAGEJSON = 'json'
F_HTML = 'html'
F_JSONLD = 'jsonld'
F_GZIP = 'gzip'
F_PNG = 'png'
F_JPEG = 'jpeg'
F_MVT = 'mvt'
F_NETCDF = 'NetCDF'

#: Formats allowed for ?f= requests (order matters for complex MIME types)
FORMAT_TYPES = OrderedDict((
    (F_HTML, 'text/html'),
    (F_JSONLD, 'application/ld+json'),
    (F_JSON, 'application/json'),
    (F_PNG, 'image/png'),
    (F_JPEG, 'image/jpeg'),
    (F_MVT, 'application/vnd.mapbox-vector-tile'),
    (F_NETCDF, 'application/x-netcdf'),
))



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2026 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

"""
Root level code of pygeoapi, parsing content provided by web framework.
Returns content from plugins and sets responses.
"""

from collections import ChainMap
from copy import deepcopy
from datetime import datetime
from functools import partial
from gzip import compress
from http import HTTPStatus
import logging
import re
import sys
from typing import Any, Tuple, Union, Self

from babel import Locale
from dateutil.parser import parse as dateparse
import pytz

from pygeoapi import __version__, l10n
from pygeoapi.api.collection import gen_collection, OGC_RELTYPES_BASE
from pygeoapi.formats import FORMAT_TYPES, F_GZIP, F_HTML, F_JSON, F_JSONLD
from pygeoapi.linked_data import jsonldify, jsonldify_collection
from pygeoapi.log import setup_logger
from pygeoapi.plugin import load_plugin
from pygeoapi.process.manager.base import get_manager
from pygeoapi.provider import filter_providers_by_type, get_provider_by_type
from pygeoapi.provider.base import ProviderGenericError, ProviderTypeError

from pygeoapi.util import (
    TEMPLATESDIR, UrlPrefetcher, filter_dict_by_key_value, get_api_rules,
    get_base_url, get_typed_value, render_j2_template, to_json,
    get_choice_from_headers, get_from_headers
)

LOGGER = logging.getLogger(__name__)

#: Return headers for requests (e.g:X-Powered-By)
HEADERS = {
    'Content-Type': 'application/json',
    'X-Powered-By': f'pygeoapi {__version__}'
}

CHARSET = ['utf-8']

#: Locale used for system responses (e.g. exceptions)
SYSTEM_LOCALE = l10n.Locale('en', 'US')

CONFORMANCE_CLASSES = [
    'http://www.opengis.net/spec/ogcapi-common-1/1.0/conf/core',
    'http://www.opengis.net/spec/ogcapi-common-2/1.0/conf/collections',
    'http://www.opengis.net/spec/ogcapi-common-1/1.0/conf/landing-page',
    'http://www.opengis.net/spec/ogcapi-common-1/1.0/conf/json',
    'http://www.opengis.net/spec/ogcapi-common-1/1.0/conf/html',
    'http://www.opengis.net/spec/ogcapi-common-1/1.0/conf/oas30'
]


def all_apis() -> dict:
    """
    Return all supported API modules

    NOTE: this is a function and not a constant to avoid import loops

    :returns: `dict` of API provider type, API module
    """

    from . import (coverages, environmental_data_retrieval, itemtypes, maps,
                   processes, pubsub, tiles, stac)

    return {
        'coverage': coverages,
        'edr': environmental_data_retrieval,
        'itemtypes': itemtypes,
        'map': maps,
        'process': processes,
        'pubsub': pubsub,
        'tile': tiles,
        'stac': stac
    }


def apply_gzip(headers: dict, content: Union[str, bytes]) -> Union[str, bytes]:
    """
    Compress content if requested in header.
    """
    charset = CHARSET[0]

    if F_GZIP in get_from_headers(headers, 'content-encoding'):
        try:
            if isinstance(content, bytes):
                # bytes means Content-Type needs to be set upstream
                content = compress(content)
            else:
                headers['Content-Type'] = \
                    f"{headers['Content-Type']}; charset={charset}"
                content = compress(content.encode(charset))
        except TypeError as err:
            headers.pop('Content-Encoding')
            LOGGER.error(f'Error in compression: {err}')
    return content


class APIRequest:
    """
    Transforms an incoming server-specific Request into an object
    with some generic helper methods and properties.

    This allows writing straightforward API functions supporting all
    web platforms such as with this example:

    .. code-block:: python

       def example_method(api: API, request: APIRequest, custom_arg):
          headers = request.get_response_headers()

          # generate response_body here

          return headers, HTTPStatus.OK, response_body


    Basic request validation is done automatically by web platform specific
    adapters such as ``execute_from_flask`` in flask_app.py . If you want to
    support custom formats (e.g. ``f=xml``), it needs to be registered with
    `skip_valid_check=True` and you can use the following code for custom
    validation. If `xml` was requested, we set the `Content-Type` ourselves.
    For the standard formats, the `APIRequest` object sets the `Content-Type`.

    .. code-block:: python

       def example_method(api: API, request: APIRequest, custom_arg):
          if not request.is_valid(['xml']):
             return api.get_format_exception(request)

          content_type = 'application/xml' if request.format == 'xml' else None
          headers = request.get_response_headers(content_type)

          # generate response_body here

          return headers, HTTPStatus.OK, response_body

    Note that you don't *have* to call :meth:`is_valid`, but that you can also
    perform a custom check on the requested output format by looking at the
    :attr:`format` property.
    Other query parameters are available through the :attr:`params` property as
    a `dict`. The request body is available through the :attr:`data` property.

    :param request:             The web platform specific Request instance.
    :param supported_locales:   List or set of supported Locale instances.
    """

    def __init__(self, request, supported_locales):
        # Set default request data
        self._data = b''

        # Copy request query parameters
        self._args = self._get_params(request)

        # Get path info
        if hasattr(request, 'scope'):
            self._path_info = request.scope['path'].strip('/')
        elif hasattr(request.headers, 'environ'):
            self._path_info = request.headers.environ['PATH_INFO'].strip('/')
        elif hasattr(request, 'path_info'):
            self._path_info = request.path_info

        # Extract locale from params or headers
        self._raw_locale, self._locale = self._get_locale(request.headers,
                                                          supported_locales)

        # Determine format
        self._format = self._get_format(request.headers)

        # Get received headers
        self._headers = self.get_request_headers(request.headers)

    @classmethod
    def from_flask(cls, request, supported_locales) -> 'APIRequest':
        """Factory class similar to with_data, but only for flask requests"""
        api_req = cls(request, supported_locales)
        api_req._data = request.data
        return api_req

    @classmethod
    async def from_starlette(cls, request, supported_locales) -> 'APIRequest':
        """Factory class similar to with_data, but only for starlette requests
        """
        api_req = cls(request, supported_locales)
        api_req._data = await request.body()
        return api_req

    @classmethod
    def from_django(cls, request, supported_locales) -> 'APIRequest':
        """Factory class similar to with_data, but only for django requests"""
        api_req = cls(request, supported_locales)
        api_req._data = request.body
        return api_req

    @staticmethod
    def _get_params(request):
        """
        Extracts the query parameters from the `Request` object.

        :param request: A Flask or Starlette Request instance
        :returns: `ImmutableMultiDict` or empty `dict`
        """

        if hasattr(request, 'args'):
            # Return ImmutableMultiDict from Flask request
            return request.args
        elif hasattr(request, 'query_params'):
            # Return ImmutableMultiDict from Starlette request
            return request.query_params
        elif hasattr(request, 'GET'):
            # Return QueryDict from Django GET request
            return request.GET
        elif hasattr(request, 'POST'):
            # Return QueryDict from Django GET request
            return request.POST
        LOGGER.debug('No query parameters found')
        return {}

    def _get_locale(self, headers: dict,
                    supported_locales: Union[list, Locale]) -> tuple:
        """
        Detects locale from "lang=<language>" param or `Accept-Language`
        header. Returns a tuple of (raw, locale) if found in params or headers.
        Returns a tuple of (raw default, default locale) if not found.

        :param headers: A dict with Request headers
        :param supported_locales: List or set of supported Locale instances
        :returns: A tuple of (str, Locale)
        """

        raw = None
        try:
            default_locale = l10n.str2locale(supported_locales[0])
        except (TypeError, IndexError, l10n.LocaleError) as err:
            # This should normally not happen, since the API class already
            # loads the supported languages from the config, which raises
            # a LocaleError if any of these languages are invalid.
            LOGGER.error(err)
            raise ValueError(f"{self.__class__.__name__} must be initialized"
                             f"with a list of valid supported locales")

        for field, mapping in ((l10n.QUERY_PARAM, self._args),
                               ('accept-language', headers)):

            loc_strs = get_choice_from_headers(mapping, field, all=True)
            if loc_strs:
                if raw is None:
                    # This is the first-found locale string: set as raw
                    raw = get_from_headers(mapping, field)

                # Check if locale string is a good match for the UI
                loc = l10n.best_match(loc_strs, supported_locales)
                is_override = field is l10n.QUERY_PARAM
                if loc != default_locale or is_override:
                    return raw, loc

        return raw, default_locale

    def _get_format(self, headers: dict,
                    extra_formats: dict = {}) -> Union[str, None]:
        """
        Get `Request` format type from query parameters or headers.

        :param headers: Dict of Request headers
        :param extra_formats: Dict of extra dataset specific formats

        :returns: format value or None if not found/specified
        """

        # Optional f=html or f=json query param
        # Overrides Accept header and might differ from FORMAT_TYPES
        format_ = (self._args.get('f') or '').strip()
        if format_:
            return format_

        # Format not specified: get from Accept headers (MIME types)
        # e.g. Accept: 'text/html;q=0.5,application/ld+json'
        types_ = get_choice_from_headers(headers, 'accept', all=True)
        if types_ is None:
            return

        merged_format_types = FORMAT_TYPES | extra_formats

        (fmts, mimes) = zip(*merged_format_types.items())
        mimes2 = [m.split(';')[0] for m in mimes]

        for type_ in types_:
            if type_ in mimes2:
                idx_ = mimes2.index(type_)
                return fmts[idx_]

    @property
    def data(self) -> bytes:
        """Returns the additional data send with the Request (bytes)"""
        return self._data

    @property
    def params(self) -> dict:
        """Returns the Request query parameters dict"""
        return self._args

    @property
    def path_info(self) -> str:
        """Returns the web server request path info part"""
        return self._path_info

    @property
    def locale(self) -> l10n.Locale:
        """
        Returns the user-defined locale from the request object.
        If no locale has been defined or if it is invalid,
        the default server locale is returned.

        .. note::   The locale here determines the language in which pygeoapi
                    should return its responses. This may not be the language
                    that the user requested. It may also not be the language
                    that is supported by a collection provider, for example.
                    For this reason, you should pass the `raw_locale` property
                    to the :func:`l10n.get_plugin_locale` function, so that
                    the best match for the provider can be determined.

        :returns: babel.core.Locale
        """

        return self._locale

    @property
    def raw_locale(self) -> Union[str, None]:
        """
        Returns the raw locale string from the `Request` object.
        If no "lang" query parameter or `Accept-Language` header was found,
        `None` is returned.
        Pass this value to the :func:`l10n.get_plugin_locale` function to let
        the provider determine a best match for the locale, which may be
        different from the locale used by pygeoapi's UI.

        :returns: a locale string or None
        """

        return self._raw_locale

    @property
    def format(self) -> Union[str, None]:
        """
        Returns the content type format from the
        request query parameters or headers.

        :returns: Format name or None
        """

        return self._format

    @property
    def headers(self) -> dict:
        """
        Returns the dictionary of the headers from
        the request.

        :returns: Request headers dictionary
        """

        return self._headers

    def get_linkrel(self, format_: str) -> str:
        """
        Returns the hyperlink relationship (rel) attribute value for
        the given API format string.

        The string is compared against the request format and if it matches,
        the value 'self' is returned. Otherwise, 'alternate' is returned.
        However, if `format_` is 'json' and *no* request format was found,
        the relationship 'self' is returned as well (JSON is the default).

        :param format_: The format to compare the request format against.
        :returns: A string 'self' or 'alternate'.
        """

        fmt = format_.lower()
        if fmt == self._format or (fmt == F_JSON and not self._format):
            return 'self'
        return 'alternate'

    def is_valid(self, additional_formats=None) -> bool:
        """
        Returns True if:
            - the format is not set (None)
            - the requested format is supported
            - the requested format exists in a list if additional formats

        .. note::   Format names are matched in a case-insensitive manner.

        :param additional_formats: Optional additional supported formats list

        :returns: bool
        """

        if not self._format:
            return True
        if self._format in FORMAT_TYPES.keys():
            return True
        if self._format in (f.lower() for f in (additional_formats or ())):
            return True
        return False

    def get_response_headers(self, force_lang: l10n.Locale | None = None,
                             force_type: str | None = None,
                             force_encoding: str | None = None,
                             **custom_headers) -> dict:
        """
        Prepares and returns a dictionary with Response object headers.

        This method always adds a 'Content-Language' header, where the value
        is determined by the 'lang' query parameter or 'Accept-Language'
        header from the request.
        If no language was requested, the default pygeoapi language is used,
        unless a `force_lang` override was specified (see notes below).

        A 'Content-Type' header is also always added to the response.
        If the user does not specify `force_type`, the header is based on
        the `format` APIRequest property. If that is invalid, the default MIME
        type `application/json` is used.

        ..note::    If a `force_lang` override is applied, that language
                    is always set as the 'Content-Language', regardless of
                    a 'lang' query parameter or 'Accept-Language' header.
                    If an API response always needs to be in the same
                    language, 'force_lang' should be set to that language.

        :param force_lang: An optional Content-Language header override.
        :param force_type: An optional Content-Type header override.
        :param force_encoding: An optional Content-Encoding header override.
        :returns: A header dict
        """

        headers = HEADERS.copy()
        headers.update(**custom_headers)
        l10n.set_response_language(headers, force_lang or self._locale)
        if force_type:
            # Set custom MIME type if specified
            headers['Content-Type'] = force_type
        elif self.is_valid() and self._format:
            # Set MIME type for valid formats
            headers['Content-Type'] = FORMAT_TYPES[self._format]

        if F_GZIP in FORMAT_TYPES:
            if force_encoding:
                headers['Content-Encoding'] = force_encoding
            elif F_GZIP in get_from_headers(self._headers, 'accept-encoding'):  # noqa
                headers['Content-Encoding'] = F_GZIP

        return headers

    def get_request_headers(self, headers: dict) -> dict:
        """
        Obtains and returns a dictionary with Request object headers.

        This method adds the headers of the original request and
        makes them available to the API object.

        :returns: A header dict
        """

        headers_ = {item[0]: item[1] for item in headers.items()}
        return headers_


class API:
    """API object"""

    def __init__(self, config: dict, openapi: dict,
                 asyncapi: dict = {}) -> Self | None:
        """
        constructor

        :param config: configuration dict
        :param openapi: openapi dict
        :param asyncapi: asyncapi dict

        :returns: `pygeoapi.API` instance
        """

        self.config = config
        self.openapi = openapi
        self.asyncapi = asyncapi
        self.api_headers = get_api_rules(self.config).response_headers
        self.base_url = get_base_url(self.config)
        self.prefetcher = UrlPrefetcher()
        self.pubsub_client = None

        CHARSET[0] = config['server'].get('encoding', 'utf-8')
        if config['server'].get('gzip'):
            FORMAT_TYPES[F_GZIP] = 'application/gzip'
            FORMAT_TYPES.move_to_end(F_JSON)

        # Process language settings (first locale is default!)
        self.locales = l10n.get_locales(config)
        self.default_locale = self.locales[0]

        if 'templates' not in self.config['server']:
            self.config['server']['templates'] = {'path': TEMPLATESDIR}

        if 'pretty_print' not in self.config['server']:
            self.config['server']['pretty_print'] = False

        self.pretty_print = self.config['server']['pretty_print']

        setup_logger(self.config['logging'])

        # Create config clone for HTML templating with modified base URL
        self.tpl_config = deepcopy(self.config)
        self.tpl_config['server']['url'] = self.base_url

        self.manager = get_manager(self.config)
        LOGGER.info('Process manager plugin loaded')

        if self.config.get('pubsub') is not None:
            LOGGER.debug('Loading PubSub client')
            self.pubsub_client = load_plugin('pubsub', self.config['pubsub'])

    def get_exception(self, status: int, headers: dict, format_: str | None,
                      code: str, description: str) -> Tuple[dict, int, str]:
        """
        Exception handler

        :param status: HTTP status code
        :param headers: dict of HTTP response headers
        :param format_: format string
        :param code: OGC API exception code
        :param description: OGC API exception code

        :returns: tuple of headers, status, and message
        """

        exception_info = sys.exc_info()

        LOGGER.error(
            description,
            exc_info=exception_info if exception_info[0] is not None else None
        )
        exception = {
            'code': code,
            'type': code,
            'description': description
        }

        if format_ == F_HTML:
            headers['Content-Type'] = FORMAT_TYPES[F_HTML]
            content = render_j2_template(
                self.tpl_config, self.config['server']['templates'],
                'exception.html', exception, SYSTEM_LOCALE)
        else:
            content = to_json(exception, self.pretty_print)

        if status == HTTPStatus.NO_CONTENT:
            LOGGER.error('HTTP 204 detected, suppressing content')
            content = ''

        return headers, status, content

    def get_format_exception(self,
                             request: APIRequest) -> Tuple[dict, int, str]:
        """
        Returns a format exception.

        :param request: An APIRequest instance.

        :returns: tuple of (headers, status, message)
        """

        # Content-Language is in the system locale (ignore language settings)
        headers = request.get_response_headers(SYSTEM_LOCALE,
                                               **self.api_headers)
        msg = 'Invalid format requested'
        LOGGER.error(f'{msg}: {request.format}')
        return self.get_exception(
            HTTPStatus.BAD_REQUEST, headers,
            request.format, 'InvalidParameterValue', msg)

    def get_collections_url(self) -> str:
        return f"{self.base_url}/collections"

    def get_dataset_templates(self, dataset: str) -> dict:
        templates = self.config['resources'][dataset].get('templates')

        return templates or self.tpl_config['server']['templates']


@jsonldify
def landing_page(api: API,
                 request: APIRequest) -> Tuple[dict, int, str]:
    """
    Provide API landing page

    :param request: A request object

    :returns: tuple of headers, status code, content
    """

    fcm = {
        'links': [],
        'title': l10n.translate(
            api.config['metadata']['identification']['title'],
            request.locale),
        'description':
            l10n.translate(
                api.config['metadata']['identification']['description'],
                request.locale)
    }

    LOGGER.debug('Creating links')
    # TODO: put title text in config or translatable files?
    fcm['links'] = [{
        'rel': 'about',
        'type': 'text/html',
        'title': l10n.translate(
            api.config['metadata']['identification']['title'],
            request.locale),
        'href': api.config['metadata']['identification']['url']
    }, {
        'rel': request.get_linkrel(F_JSON),
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('This document as JSON', request.locale),
        'href': f"{api.base_url}?f={F_JSON}"
    }, {
        'rel': request.get_linkrel(F_JSONLD),
        'type': FORMAT_TYPES[F_JSONLD],
        'title': l10n.translate('This document as RDF (JSON-LD)', request.locale),  # noqa
        'href': f"{api.base_url}?f={F_JSONLD}"
    }, {
        'rel': request.get_linkrel(F_HTML),
        'type': FORMAT_TYPES[F_HTML],
        'title': l10n.translate('This document as HTML', request.locale),
        'href': f"{api.base_url}?f={F_HTML}",
        'hreflang': api.default_locale
    }, {
        'rel': 'service-desc',
        'type': 'application/vnd.oai.openapi+json;version=3.0',
        'title': l10n.translate('The OpenAPI definition as JSON', request.locale),  # noqa
        'href': f"{api.base_url}/openapi"
    }, {
        'rel': 'service-doc',
        'type': FORMAT_TYPES[F_HTML],
        'title': l10n.translate('The OpenAPI definition as HTML', request.locale),  # noqa
        'href': f"{api.base_url}/openapi?f={F_HTML}",
        'hreflang': api.default_locale
    }, {
        'rel': 'conformance',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('Conformance', request.locale),
        'href': f"{api.base_url}/conformance"
    }, {
        'rel': 'data',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('Collections', request.locale),
        'href': api.get_collections_url()
    }, {
        'rel': f'{OGC_RELTYPES_BASE}/processes',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('Processes', request.locale),
        'href': f"{api.base_url}/processes"
    }, {
        'rel': f'{OGC_RELTYPES_BASE}/job-list',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('Jobs', request.locale),
        'href': f"{api.base_url}/jobs"
    }, {
        'rel': f'{OGC_RELTYPES_BASE}/tiling-schemes',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('The list of supported tiling schemes as JSON', request.locale),  # noqa
        'href': f"{api.base_url}/TileMatrixSets?f=json"
    }, {
        'rel': f'{OGC_RELTYPES_BASE}/tiling-schemes',
        'type': FORMAT_TYPES[F_HTML],
        'title': l10n.translate('The list of supported tiling schemes as HTML', request.locale),  # noqa
        'href': f"{api.base_url}/TileMatrixSets?f=html"
    }]

    if api.pubsub_client is not None and not api.pubsub_client.hidden:
        LOGGER.debug('Adding PubSub broker link')
        pubsub_link = {
            'rel': 'hub',
            'type': 'application/json',
            'title': l10n.translate('Pub/Sub broker', request.locale),
            'href': api.pubsub_client.broker_safe_url
        }
        if api.pubsub_client.channel is not None:
            pubsub_link['channel'] = api.pubsub_client.channel

        fcm['links'].append(pubsub_link)

    if api.asyncapi:
        fcm['links'].append({
            'rel': 'service-doc',
            'type': 'text/html',
            'title': l10n.translate('The AsyncAPI definition as HTML', request.locale),  # noqa
            'href': f'{api.base_url}/asyncapi?f=html'
        })
        fcm['links'].append({
            'rel': 'service-desc',
            'type': 'application/asyncapi+json',
            'title': l10n.translate('The AsyncAPI definition as JSON', request.locale),  # noqa
            'href': f'{api.base_url}/asyncapi?f=json'
        })

    headers = request.get_response_headers(**api.api_headers)
    if request.format == F_HTML:  # render

        for resource_type in ['collection', 'process', 'stac-collection']:
            fcm[resource_type] = False

            found = filter_dict_by_key_value(api.config['resources'],
                                             'type', resource_type)
            if found:
                fcm[resource_type] = True
                if resource_type == 'collection':  # check for tiles
                    for key, value in found.items():
                        if filter_providers_by_type(value['providers'],
                                                    'tile'):
                            fcm['tile'] = True

        if api.pubsub_client is not None and not api.pubsub_client.hidden:
            fcm['pubsub'] = {
                'name': api.pubsub_client.name,
                'url': api.pubsub_client.broker_safe_url,
                'channel': api.pubsub_client.channel,
                'asyncapi': api.asyncapi
            }

        content = render_j2_template(
            api.tpl_config, api.config['server']['templates'],
            'landing_page.html', fcm, request.locale)

        return headers, HTTPStatus.OK, content

    if request.format == F_JSONLD:
        return headers, HTTPStatus.OK, to_json(
            api.fcmld, api.pretty_print)

    return headers, HTTPStatus.OK, to_json(fcm, api.pretty_print)


def openapi_(api: API, request: APIRequest) -> Tuple[dict, int, str]:
    """
    Provide OpenAPI document

    :param request: A request object
    :param openapi: dict of OpenAPI definition

    :returns: tuple of headers, status code, content
    """
    headers = request.get_response_headers(**api.api_headers)

    if request.format == F_HTML:
        template = 'openapi/swagger.html'
        if request._args.get('ui') == 'redoc':
            template = 'openapi/redoc.html'

        path = f'{api.base_url}/openapi'
        data = {
            'openapi-document-path': path
        }
        content = render_j2_template(
            api.tpl_config, api.config['server']['templates'], template, data,
            request.locale)

        return headers, HTTPStatus.OK, content

    headers['Content-Type'] = 'application/vnd.oai.openapi+json;version=3.0'  # noqa

    if isinstance(api.openapi, dict):
        return headers, HTTPStatus.OK, to_json(api.openapi,
                                               api.pretty_print)
    else:
        return headers, HTTPStatus.OK, api.openapi


def asyncapi_(api: API, request: APIRequest) -> Tuple[dict, int, str]:
    """
    Provide AsyncAPI document

    :param request: A request object

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers(**api.api_headers)

    if not api.asyncapi:
        msg = 'AsyncAPI not supported/configured'
        return api.get_exception(
            HTTPStatus.NOT_IMPLEMENTED, headers, request.format,
            'NoApplicableCode', msg)

    if request.format == F_HTML:
        template = 'asyncapi.html'

        path = f'{api.base_url}/asyncapi'
        data = {
            'asyncapi-document-path': path
        }
        content = render_j2_template(
            api.tpl_config, api.config['server']['templates'], template, data,
            request.locale)

        return headers, HTTPStatus.OK, content

    headers['Content-Type'] = 'application/asyncapi+json'

    return headers, HTTPStatus.OK, to_json(api.asyncapi, api.pretty_print)


def conformance(api: API, request: APIRequest) -> Tuple[dict, int, str]:
    """
    Provide conformance definition

    :param request: A request object

    :returns: tuple of headers, status code, content
    """

    apis_dict = all_apis()

    conformance_list = list(CONFORMANCE_CLASSES)

    for key, value in api.config['resources'].items():
        if value['type'] == 'process':
            conformance_list.extend(
                apis_dict['process'].CONFORMANCE_CLASSES)
        else:
            for provider in value['providers']:
                if provider['type'] in apis_dict:
                    conformance_list.extend(
                        apis_dict[provider['type']].CONFORMANCE_CLASSES)
                if provider['type'] == 'feature':
                    conformance_list.extend(
                        apis_dict['itemtypes'].CONFORMANCE_CLASSES_FEATURES)  # noqa
                if provider['type'] == 'record':
                    conformance_list.extend(
                        apis_dict['itemtypes'].CONFORMANCE_CLASSES_RECORDS)

    if api.pubsub_client is not None:
        conformance_list.extend(apis_dict['pubsub'].CONFORMANCE_CLASSES)

    conformance = {
        'conformsTo': sorted(list(set(conformance_list)))
    }

    headers = request.get_response_headers(**api.api_headers)
    if request.format == F_HTML:  # render
        content = render_j2_template(
            api.tpl_config, api.config['server']['templates'],
            'conformance.html', conformance, request.locale)

        return headers, HTTPStatus.OK, content

    return headers, HTTPStatus.OK, to_json(conformance, api.pretty_print)


@jsonldify
def describe_collections(api: API, request: APIRequest,
                         dataset: str | None = None) -> Tuple[dict, int, str]:
    """
    Provide collection metadata

    :param request: A request object
    :param dataset: name of collection

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers(**api.api_headers)

    fcm = {
        'collections': [],
        'links': []
    }

    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'collection')

    if all([dataset is not None, dataset not in collections.keys()]):
        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    if dataset is not None:
        collections_dict = {dataset: api.config['resources'][dataset]}
    else:
        collections_dict = collections

    LOGGER.debug('Creating collections')
    for k, v in collections_dict.items():
        if v.get('visibility', 'default') == 'hidden':
            LOGGER.debug(f'Skipping hidden layer: {k}')
            continue

        try:
            fcm['collections'].append(
                gen_collection(api, request, k, request.locale))
        except Exception as err:
            LOGGER.warning(f'Error generating collection {k}: {err}')
            if dataset is None:
                LOGGER.debug('Skipping failed dataset')
            else:
                return api.get_exception(
                    HTTPStatus.INTERNAL_SERVER_ERROR, headers, request.format,
                    'NoApplicableCode', 'Error generating collection')

    if dataset is not None:
        fcm = fcm['collections'][0]

    if dataset is None:
        # TODO: translate
        fcm['links'].append({
            'type': FORMAT_TYPES[F_JSON],
            'rel': request.get_linkrel(F_JSON),
            'title': l10n.translate('This document as JSON', request.locale),  # noqa
            'href': f'{api.get_collections_url()}?f={F_JSON}'
        })
        fcm['links'].append({
            'type': FORMAT_TYPES[F_JSONLD],
            'rel': request.get_linkrel(F_JSONLD),
            'title': l10n.translate('This document as RDF (JSON-LD)', request.locale),  # noqa
            'href': f'{api.get_collections_url()}?f={F_JSONLD}'
        })
        fcm['links'].append({
            'type': FORMAT_TYPES[F_HTML],
            'rel': request.get_linkrel(F_HTML),
            'title': l10n.translate('This document as HTML', request.locale),  # noqa
            'href': f'{api.get_collections_url()}?f={F_HTML}'
        })

    if request.format == F_HTML:  # render
        fcm['base_url'] = api.base_url
        fcm['collections_path'] = api.get_collections_url()
        if dataset is not None:
            tpl_config = api.get_dataset_templates(dataset)
            content = render_j2_template(api.tpl_config, tpl_config,
                                         'collections/collection.html',
                                         fcm, request.locale)
        else:
            content = render_j2_template(
                api.tpl_config, api.config['server']['templates'],
                'collections/index.html', fcm, request.locale)

        return headers, HTTPStatus.OK, content

    if request.format == F_JSONLD:
        jsonld = api.fcmld.copy()
        if dataset is not None:
            jsonld['dataset'] = jsonldify_collection(api, fcm,
                                                     request.locale)
        else:
            jsonld['dataset'] = [
                jsonldify_collection(api, c, request.locale)
                for c in fcm.get('collections', [])
            ]
        return headers, HTTPStatus.OK, to_json(jsonld, api.pretty_print)

    return headers, HTTPStatus.OK, to_json(fcm, api.pretty_print)


def get_collection_schema(api: API, request: Union[APIRequest, Any],
                          dataset: str) -> Tuple[dict, int, str]:
    """
    Returns a collection schema

    :param request: A request object
    :param dataset: dataset name

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers(**api.api_headers)

    if any([dataset is None,
            dataset not in api.config['resources'].keys()]):

        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    LOGGER.debug('Creating collection schema')
    try:
        LOGGER.debug('Loading feature provider')
        p = load_plugin('provider', get_provider_by_type(
            api.config['resources'][dataset]['providers'], 'feature'))
    except ProviderTypeError:
        try:
            LOGGER.debug('Loading coverage provider')
            p = load_plugin('provider', get_provider_by_type(
                api.config['resources'][dataset]['providers'], 'coverage'))  # noqa
        except ProviderTypeError:
            try:
                LOGGER.debug('Loading record provider')
                p = load_plugin('provider', get_provider_by_type(
                    api.config['resources'][dataset]['providers'], 'record'))
            except ProviderTypeError:
                LOGGER.debug('Loading edr provider')
                p = load_plugin('provider', get_provider_by_type(
                    api.config['resources'][dataset]['providers'], 'edr'))
    except ProviderGenericError as err:
        LOGGER.error(err)
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    schema = {
        'type': 'object',
        'title': l10n.translate(
            api.config['resources'][dataset]['title'], request.locale),
        'properties': {},
        '$schema': 'http://json-schema.org/draft/2019-09/schema',
        '$id': f'{api.get_collections_url()}/{dataset}/schema'
    }

    if p.type != 'coverage':
        schema['properties']['geometry'] = {
            'format': 'geometry-any',
            'x-ogc-role': 'primary-geometry'
        }

    for k, v in p.fields.items():
        schema['properties'][k] = v
        if v['type'] == 'float':
            schema['properties'][k]['type'] = 'number'
        if v.get('format') is None:
            schema['properties'][k].pop('format', None)

        if k == p.id_field:
            schema['properties'][k]['x-ogc-role'] = 'id'
        if k == p.time_field:
            schema['properties'][k]['x-ogc-role'] = 'primary-instant'

    if request.format == F_HTML:  # render
        tpl_config = api.get_dataset_templates(dataset)
        schema['title'] = l10n.translate(
            api.config['resources'][dataset]['title'], request.locale)

        schema['collections_path'] = api.get_collections_url()
        schema['dataset_path'] = f'{api.get_collections_url()}/{dataset}'

        content = render_j2_template(api.tpl_config, tpl_config,
                                     'collections/schema.html',
                                     schema, request.locale)

        return headers, HTTPStatus.OK, content

    headers['Content-Type'] = 'application/schema+json'

    return headers, HTTPStatus.OK, to_json(schema, api.pretty_print)


def validate_bbox(value: list | None = None) -> list:
    """
    Helper function to validate bbox parameter

    :param value: `list` of minx, miny, maxx, maxy

    :returns: bbox as `list` of `float` values
    """

    if value is None:
        LOGGER.debug('bbox is empty')
        return []

    bbox = value.split(',')

    if len(bbox) not in [4, 6]:
        msg = 'bbox should be either 4 values (minx,miny,maxx,maxy) ' \
              'or 6 values (minx,miny,minz,maxx,maxy,maxz)'
        LOGGER.debug(msg)
        raise ValueError(msg)

    try:
        bbox = [float(c) for c in bbox]
    except ValueError as err:
        msg = 'bbox values must be numbers'
        err.args = (msg,)
        LOGGER.debug(msg)
        raise

    if (len(bbox) == 4 and bbox[1] > bbox[3]) \
            or (len(bbox) == 6 and bbox[1] > bbox[4]):
        msg = 'miny should be less than maxy'
        LOGGER.debug(msg)
        raise ValueError(msg)

    if (len(bbox) == 4 and bbox[0] > bbox[2]) \
            or (len(bbox) == 6 and bbox[0] > bbox[3]):
        msg = 'minx is greater than maxx (possibly antimeridian bbox)'
        LOGGER.debug(msg)

    if len(bbox) == 6 and bbox[2] > bbox[5]:
        msg = 'minz should be less than maxz'
        LOGGER.debug(msg)
        raise ValueError(msg)

    return bbox


def validate_datetime(resource_def: dict,
                      datetime_: str | None = None) -> str:
    """
    Helper function to validate temporal parameter

    :param resource_def: `dict` of configuration resource definition
    :param datetime_: `str` of datetime parameter

    :returns: `str` of datetime input, if valid
    """

    # TODO: pass datetime to query as a `datetime` object
    # we would need to ensure partial dates work accordingly
    # as well as setting '..' values to `None` so that underlying
    # providers can just assume a `datetime.datetime` object
    #
    # NOTE: needs testing when passing partials from API to backend

    datetime_invalid = False

    if datetime_ is not None and 'temporal' in resource_def:

        dateparse_begin = partial(dateparse, default=datetime.min)
        dateparse_end = partial(dateparse, default=datetime.max)
        unix_epoch = datetime(1970, 1, 1, 0, 0, 0)
        dateparse_ = partial(dateparse, default=unix_epoch)

        te = resource_def['temporal']

        try:
            if te['begin'] is not None and te['begin'].tzinfo is None:
                te['begin'] = te['begin'].replace(tzinfo=pytz.UTC)
            if te['end'] is not None and te['end'].tzinfo is None:
                te['end'] = te['end'].replace(tzinfo=pytz.UTC)
        except AttributeError:
            msg = 'Configured times should be RFC3339'
            LOGGER.error(msg)
            raise ValueError(msg)

        if '/' in datetime_:  # envelope
            LOGGER.debug('detected time range')
            LOGGER.debug('Validating time windows')

            # normalize "" to ".." (actually changes datetime_)
            datetime_ = re.sub(r'^/', '../', datetime_)
            datetime_ = re.sub(r'/$', '/..', datetime_)

            datetime_begin, datetime_end = datetime_.split('/')
            if datetime_begin != '..':
                datetime_begin = dateparse_begin(datetime_begin)
                if datetime_begin.tzinfo is None:
                    datetime_begin = datetime_begin.replace(
                        tzinfo=pytz.UTC)

            if datetime_end != '..':
                datetime_end = dateparse_end(datetime_end)
                if datetime_end.tzinfo is None:
                    datetime_end = datetime_end.replace(tzinfo=pytz.UTC)

            datetime_invalid = any([
                (te['end'] is not None and datetime_begin != '..' and
                    datetime_begin > te['end']),
                (te['begin'] is not None and datetime_end != '..' and
                    datetime_end < te['begin'])
            ])

        else:  # time instant
            LOGGER.debug('detected time instant')
            datetime__ = dateparse_(datetime_)
            if datetime__ != '..':
                if datetime__.tzinfo is None:
                    datetime__ = datetime__.replace(tzinfo=pytz.UTC)
            datetime_invalid = any([
                (te['begin'] is not None and datetime__ != '..' and
                    datetime__ < te['begin']),
                (te['end'] is not None and datetime__ != '..' and
                    datetime__ > te['end'])
            ])

    if datetime_invalid:
        msg = 'datetime parameter out of range'
        LOGGER.debug(msg)
        raise ValueError(msg)

    return datetime_


def validate_subset(value: str) -> dict:
    """
    Helper function to validate subset parameter

    :param value: `subset` parameter

    :returns: dict of axis/values
    """

    subsets = {}

    for s in value.split(','):
        LOGGER.debug(f'Processing subset {s}')
        m = re.search(r'(.*)\((.*)\)', s)
        subset_name, values = m.group(1, 2)

        if '"' in values:
            LOGGER.debug('Values are strings')
            if values.count('"') % 2 != 0:
                msg = 'Invalid format: subset should be like axis("min"[:"max"])'  # noqa
                LOGGER.error(msg)
                raise ValueError(msg)
            try:
                LOGGER.debug('Value is an interval')
                m = re.search(r'"(\S+)":"(\S+)"', values)
                values = list(m.group(1, 2))
            except AttributeError:
                LOGGER.debug('Value is point')
                m = re.search(r'"(.*)"', values)
                values = [m.group(1)]
        else:
            LOGGER.debug('Values are numbers')
            try:
                LOGGER.debug('Value is an interval')
                m = re.search(r'(\S+):(\S+)', values)
                values = list(m.group(1, 2))
            except AttributeError:
                LOGGER.debug('Value is point')
                values = [values]

        subsets[subset_name] = list(map(get_typed_value, values))

    return subsets


def evaluate_limit(requested: Union[None, int], server_limits: dict,
                   collection_limits: dict) -> int:
    """
    Helper function to evaluate limit parameter

    :param requested: the limit requested by the client
    :param server_limits: `dict` of server limits
    :param collection_limits: `dict` of collection limits

    :returns: `int` of evaluated limit
    """

    effective_limits = ChainMap(collection_limits, server_limits)

    default = effective_limits.get('default_items', 10)
    max_ = effective_limits.get('max_items', 10)
    on_exceed = effective_limits.get('on_exceed', 'throttle')

    LOGGER.debug(f'Requested limit: {requested}')
    LOGGER.debug(f'Default limit: {default}')
    LOGGER.debug(f'Maximum limit: {max_}')
    LOGGER.debug(f'On exceed: {on_exceed}')

    if requested is None:
        LOGGER.debug('no limit requested; returning default')
        return default

    if isinstance(requested, int):
        requested2 = requested
    else:
        requested2 = get_typed_value(requested)

    if not isinstance(requested2, int):
        raise ValueError('limit value should be an integer')

    if requested2 <= 0:
        raise ValueError('limit value should be strictly positive')
    elif requested2 > max_ and on_exceed == 'error':
        raise ValueError(f'Limit of {max_} exceeded')
    else:
        LOGGER.debug('limit requested')
        return min(requested2, max_)



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
import random
from urllib.parse import urlparse

from pygeoapi.error import GenericError
from pygeoapi.util import remove_url_auth

LOGGER = logging.getLogger(__name__)


class BasePubSubClient:
    """Base Pub/Sub client"""

    def __init__(self, publisher_def: dict):
        """
        Initialize object

        :param publisher_def: publisher definition

        :returns: pycsw.broker.base.BasePubSubClient
        """

        self.type = 'pubsub'

        try:
            self.name = publisher_def['name']
            self.broker = publisher_def['broker']['url']
        except KeyError:
            raise RuntimeError('name/type/broker.url are required')

        self.broker_url = urlparse(self.broker)
        self.broker_safe_url = remove_url_auth(self.broker)

        self.hidden = publisher_def['broker'].get('hidden', False)
        self.channel = publisher_def['broker'].get('channel')
        self.client_id = f'pygeoapi-pubsub-{random.randint(0, 1000)}'

    def connect(self) -> None:
        """
        Connect to a Pub/Sub broker

        :returns: None
        """

        raise NotImplementedError()

    def pub(self, channel: str, message: str) -> bool:
        """
        Publish a message to a broker/channel

        :param channel: `str` of channel
        :param message: `str` of message

        :returns: `bool` of publish result
        """

        raise NotImplementedError()

    def __repr__(self):
        return f'<BasePubSubClient> {self.broker_safe_url}'


class PubSubClientConnectionError(GenericError):
    """Pub/Sub client client connection error"""

    default_msg = 'Pub/Sub client connection error (check logs)'


class PubSubClientSubscriptionError(GenericError):
    """Pub/Sub client client subscription error"""

    default_msg = 'Pub/Sub client subscription error (check logs)'


class PubSubClientPublishError(GenericError):
    """Pub/Sub client client publish error"""

    default_msg = 'Pub/Sub client publish error (check logs)'
